# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import os
from typing import Set, List

from nlp_architect.common.cdc.mention_data import MentionDataLight
from nlp_architect.data.cdc_resources.data_types.wn.wordnet_page import WordnetPage
from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.data.cdc_resources.relations.relation_types_enums import RelationType, \
    OnlineOROfflineMethod
from nlp_architect.data.cdc_resources.wordnet.wordnet_offline import WordnetOffline
from nlp_architect.data.cdc_resources.wordnet.wordnet_online import WordnetOnline
from nlp_architect.utils.string_utils import StringUtils

logger = logging.getLogger(__name__)


class WordnetRelationExtraction(RelationExtraction):
    def __init__(self, method: OnlineOROfflineMethod = OnlineOROfflineMethod.ONLINE,
                 wn_file: str = None):
        """
        Extract Relation between two mentions according to Word Embedding cosine distance

        Args:
            method (required): OnlineOROfflineMethod.{ONLINE/OFFLINE} run against full wordnet or
                a sub-set of it (default = ONLINE)
            wn_file (required on OFFLINE mode): str Location of wordnet subset file to work with
        """
        logger.info('Loading Wordnet module')
        self.connectivity = method
        if self.connectivity == OnlineOROfflineMethod.ONLINE:
            self.wordnet_impl = WordnetOnline()
        elif self.connectivity == OnlineOROfflineMethod.OFFLINE:
            if wn_file is not None and os.path.isdir(wn_file):
                self.wordnet_impl = WordnetOffline(wn_file)
            else:
                raise FileNotFoundError('WordNet resource directory not found or not in path')

        logger.info('Wordnet module lead successfully')
        super(WordnetRelationExtraction, self).__init__()

    def extract_all_relations(self, mention_x: MentionDataLight,
                              mention_y: MentionDataLight) -> Set[RelationType]:
        """
        Try to find if mentions has anyone or more of the relations this class support

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            Set[RelationType]: One or more of: RelationType.WORDNET_SAME_SYNSET_ENTITY,
                RelationType.WORDNET_SAME_SYNSET_EVENT, RelationType.WORDNET_PARTIAL_SYNSET_MATCH,
                RelationType.WORDNET_DERIVATIONALLY
        """
        relations = set()
        mention_x_str = mention_x.tokens_str
        mention_y_str = mention_y.tokens_str
        if StringUtils.is_pronoun(mention_x_str.lower()) or StringUtils.is_pronoun(
                mention_y_str.lower()):
            relations.add(RelationType.NO_RELATION_FOUND)
            return relations

        page_x = self.wordnet_impl.get_pages(mention_x)
        page_y = self.wordnet_impl.get_pages(mention_y)

        if page_x and page_y:
            deriv_rel = self.extract_derivation(page_x, page_y)
            part_syn_rel = self.extract_partial_synset_match(page_x, page_y)
            same_syn_rel = self.extract_same_synset_entity(page_x, page_y)
            if deriv_rel != RelationType.NO_RELATION_FOUND:
                relations.add(deriv_rel)
            if part_syn_rel != RelationType.NO_RELATION_FOUND:
                relations.add(part_syn_rel)
            if same_syn_rel != RelationType.NO_RELATION_FOUND:
                relations.add(same_syn_rel)

        if len(relations) == 0:
            relations.add(RelationType.NO_RELATION_FOUND)

        return relations

    def extract_sub_relations(self, mention_x: MentionDataLight, mention_y: MentionDataLight,
                              relation: RelationType) -> RelationType:
        """
        Check if input mentions has the given relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight
            relation: RelationType

        Returns:
            RelationType: relation in case mentions has given relation or
                RelationType.NO_RELATION_FOUND otherwise
        """
        mention_x_str = mention_x.tokens_str
        mention_y_str = mention_y.tokens_str
        if StringUtils.is_pronoun(mention_x_str.lower()) or StringUtils.is_pronoun(
                mention_y_str.lower()):
            return RelationType.NO_RELATION_FOUND

        page_x = self.wordnet_impl.get_pages(mention_x)
        page_y = self.wordnet_impl.get_pages(mention_y)

        if page_x and page_y:
            if relation == RelationType.WORDNET_DERIVATIONALLY:
                return self.extract_derivation(page_x, page_y)
            if relation == RelationType.WORDNET_PARTIAL_SYNSET_MATCH:
                return self.extract_partial_synset_match(page_x, page_y)
            if relation == RelationType.WORDNET_SAME_SYNSET:
                return self.extract_same_synset_entity(page_x, page_y)

        return RelationType.NO_RELATION_FOUND

    @staticmethod
    def extract_derivation(page_x: WordnetPage, page_y: WordnetPage) -> RelationType:
        """
        Check if input mentions has derivation relation

        Args:
            page_x:WordnetPage
            page_y:WordnetPage

        Returns:
            RelationType.WORDNET_DERIVATIONALLY or RelationType.NO_RELATION_FOUND
        """
        x_head = page_x.head
        x_head_lemma = page_x.head_lemma
        y_head = page_y.head
        y_head_lemma = page_y.head_lemma

        x_set = set()
        x_set.update(page_x.head_derivationally)
        x_set.update(page_x.head_lemma_derivationally)

        y_set = set()
        y_set.update(page_y.head_derivationally)
        y_set.update(page_y.head_lemma_derivationally)

        relation = RelationType.NO_RELATION_FOUND

        if y_head in x_set or y_head_lemma in x_set or x_head in y_set or \
                x_head_lemma in y_set or len(x_set & y_set) > 0:
            relation = RelationType.WORDNET_DERIVATIONALLY
            # print 'matched by derivation - ' + str(x_head)+ ' , ' + str(y_head)

        return relation

    @staticmethod
    def extract_partial_synset_match(page_x: WordnetPage, page_y: WordnetPage) -> RelationType:
        """
        Check if input mentions has partial synset relation

        Args:
            page_x:WordnetPage
            page_y:WordnetPage

        Returns:
            RelationType.WORDNET_PARTIAL_SYNSET_MATCH or RelationType.NO_RELATION_FOUND
        """
        x_words = page_x.clean_phrase.split()
        y_words = page_y.clean_phrase.split()

        if len(x_words) == 0 or len(y_words) == 0:
            return RelationType.NO_RELATION_FOUND

        x_synonyms = page_x.all_clean_words_synonyms
        y_synonyms = page_y.all_clean_words_synonyms

        # One word - check whether there is intersection between synsets
        if len(x_synonyms) == 1 and len(y_synonyms) == 1 and \
                len([w for w in (x_synonyms[0] & y_synonyms[0])]) > 0:
            # print 'matched by partial - ' + str(y) + ' , ' + str(x)
            return RelationType.WORDNET_PARTIAL_SYNSET_MATCH

        return RelationType.NO_RELATION_FOUND

    @staticmethod
    def extract_same_synset_entity(page_x: WordnetPage, page_y: WordnetPage) -> RelationType:
        """
        Check if input mentions has same synset relation for entity mentions

        Args:
            page_x:WordnetPage
            page_y:WordnetPage

        Returns:
            RelationType.WORDNET_SAME_SYNSET_ENTITY or RelationType.NO_RELATION_FOUND
        """
        match_result = RelationType.NO_RELATION_FOUND
        th = 0
        if len([w for w in (page_x.head_synonyms & page_y.head_synonyms)]) > th:
            match_result = RelationType.WORDNET_SAME_SYNSET

        return match_result

    @staticmethod
    def get_supported_relations() -> List[RelationType]:
        """
        Return all supported relations by this class

        Returns:
            List[RelationType]
        """
        return [RelationType.WORDNET_SAME_SYNSET,
                RelationType.WORDNET_PARTIAL_SYNSET_MATCH, RelationType.WORDNET_DERIVATIONALLY]



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import re
from typing import Dict

from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_page_extracted_relations import \
    WikipediaPageExtractedRelations, DISAMBIGUATION_TITLE
from nlp_architect.utils.string_utils import StringUtils


class WikipediaPage(object):
    def __init__(self, orig_phrase: str = None, orig_phrase_norm: str = None,
                 wiki_title: str = None, wiki_title_norm: str = None,
                 score: int = 0, pageid: int = 0, description: str = None,
                 relations: WikipediaPageExtractedRelations = None) -> None:
        """
        Object represent a Wikipedia Page and extracted fields.

        Args:
            orig_phrase (str): original search phrase
            orig_phrase_norm (str): original search phrase normalized
            wiki_title (str): page title
            wiki_title_norm (str): page title normalized
            score (int): score for getting wiki_title from orig_phrase
            pageid (int): the unique page identifier
            description (str, optional): the page description
            relations (WikipediaPageExtractedRelations): Object that represent all
                                                         extracted Wikipedia relations
        """
        self.orig_phrase = orig_phrase
        if orig_phrase_norm is None:
            self.orig_phrase_norm = StringUtils.normalize_str(orig_phrase)
        else:
            self.orig_phrase_norm = orig_phrase_norm

        self.wiki_title = wiki_title.replace(DISAMBIGUATION_TITLE, '')
        if wiki_title_norm is None:
            self.wiki_title_norm = StringUtils.normalize_str(wiki_title)
        else:
            self.wiki_title_norm = wiki_title_norm

        self.score = score
        self.pageid = int(pageid)
        self.description = description
        self.relations = relations

    def toJson(self) -> Dict:
        result_dict = {}
        result_dict['orig_phrase'] = self.orig_phrase
        result_dict['orig_phrase_norm'] = self.orig_phrase_norm
        result_dict['wiki_title'] = self.wiki_title
        result_dict['wiki_title_norm'] = self.wiki_title_norm
        result_dict['score'] = self.score
        result_dict['pageid'] = self.pageid
        result_dict['description'] = self.description
        result_dict['relations'] = self.relations.toJson()
        return result_dict

    def __eq__(self, other):
        return self.orig_phrase == other.orig_phrase and self.wiki_title == other.wiki_title and \
            self.pageid == other.pageid

    def __hash__(self):
        return hash(self.orig_phrase) + hash(self.pageid) + hash(self.wiki_title)

    def __str__(self) -> str:
        result_str = ''
        try:
            title_strip = re.sub(u'(\u2018|\u2019)', '\'', self.orig_phrase)
            wiki_title_strip = re.sub(u'(\u2018|\u2019)', '\'', self.wiki_title)
            result_str = str(title_strip) + ', ' + str(wiki_title_strip) + ', ' + \
                str(self.score) + ', ' + str(self.pageid) + ', ' + \
                str(self.description) + ', ' + str(self.relations)
        except Exception:
            result_str = 'error in to_string()'

        return result_str



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import division, print_function, unicode_literals, absolute_import

from os import path

import numpy as np

from nlp_architect.utils.generic import pad_sentences
from nlp_architect.utils.io import validate_existing_directory, validate_existing_filepath
from nlp_architect.utils.text import read_sequential_tagging_file, \
    word_vector_generator, character_vector_generator


class SequentialTaggingDataset(object):
    """
    Sequential tagging dataset loader.
    Loads train/test files with tabular separation.

    Args:
        train_file (str): path to train file
        test_file (str): path to test file
        max_sentence_length (int, optional): max sentence length
        max_word_length (int, optional): max word length
        tag_field_no (int, optional): index of column to use a y-samples
    """

    def __init__(self,
                 train_file,
                 test_file,
                 max_sentence_length=30,
                 max_word_length=20,
                 tag_field_no=2):
        self.files = {'train': train_file,
                      'test': test_file}
        self.max_sent_len = max_sentence_length
        self.max_word_len = max_word_length
        self.tf = tag_field_no

        self.vocabs = {'token': None,
                       'char': None,   # 0=pad, 1=unk
                       'tag': None}    # 0=pad

        self.data = {}

        sentences = self._read_file(self.files['train'])
        train_size = len(sentences)
        sentences += self._read_file(self.files['test'])
        test_size = len(sentences) - train_size
        texts, tags = list(zip(*sentences))

        texts_mat, self.vocabs['token'] = word_vector_generator(texts, lower=True, start=2)
        tags_mat, self.vocabs['tag'] = word_vector_generator(tags, start=1)
        chars_mat, self.vocabs['char'] = character_vector_generator(texts, start=2)

        texts_mat = pad_sentences(texts_mat, max_length=self.max_sent_len)
        tags_mat = pad_sentences(tags_mat, max_length=self.max_sent_len)

        chars_mat = [pad_sentences(d, max_length=self.max_word_len) for d in chars_mat]
        zeros = np.zeros((len(chars_mat), self.max_sent_len, self.max_word_len))
        for idx, d in enumerate(chars_mat):
            d = d[:self.max_sent_len]
            zeros[idx, :d.shape[0]] = d
        chars_mat = zeros.astype(dtype=np.int32)

        self.data['train'] = texts_mat[:train_size], chars_mat[:train_size], tags_mat[:train_size]
        self.data['test'] = texts_mat[-test_size:], chars_mat[-test_size:], tags_mat[-test_size:]

    @property
    def y_labels(self):
        """return y labels"""
        return self.vocabs['tag']

    @property
    def word_vocab(self):
        """words vocabulary"""
        return self.vocabs['token']

    @property
    def char_vocab(self):
        """characters vocabulary"""
        return self.vocabs['char']

    @property
    def word_vocab_size(self):
        """word vocabulary size"""
        return len(self.vocabs['token']) + 2

    @property
    def char_vocab_size(self):
        """character vocabulary size"""
        return len(self.vocabs['char']) + 2

    @property
    def train_set(self):
        """Get the train set"""
        return self.data['train']

    @property
    def test_set(self):
        """Get the test set"""
        return self.data['test']

    def _read_file(self, filepath):
        with open(filepath, encoding='utf-8') as fp:
            data = fp.readlines()
            data = [d.strip() for d in data]
            sentences = self._split_into_sentences(data)
            parsed_sentences = [self._parse_sentence(s) for s in sentences if len(s) > 0]
        return parsed_sentences

    def _parse_sentence(self, sentence):
        tokens = []
        tags = []
        for line in sentence:
            fields = line.split()
            assert len(fields) >= self.tf, 'tag field exceeds number of fields'
            if 'CD' in fields[1]:
                tokens.append('0')
            else:
                tokens.append(fields[0])
            tags.append(fields[self.tf - 1])
        return tokens, tags

    @staticmethod
    def _split_into_sentences(file_lines):
        sents = []
        s = []
        for line in file_lines:
            line = line.strip()
            if not line:
                sents.append(s)
                s = []
                continue
            s.append(line)
        sents.append(s)
        return sents


class CONLL2000(object):
    """
        CONLL 2000 POS/chunking task data set (numpy)

        Arguments:
            data_path (str): directory containing CONLL2000 files
            sentence_length (int, optional): number of time steps to embed the data.
                None value will not truncate vectors
            max_word_length (int, optional): max word length in characters.
                None value will not truncate vectors
            extract_chars (boolean, optional): Yield Char RNN features.
            lowercase (bool, optional): lower case sentence words
        """

    dataset_files = {'train': 'train.txt',
                     'test': 'test.txt'}

    def __init__(self,
                 data_path,
                 sentence_length=None,
                 max_word_length=None,
                 extract_chars=False,
                 lowercase=True):
        self._validate_paths(data_path)
        self.data_path = data_path
        self.sentence_length = sentence_length
        self.use_chars = extract_chars
        self.max_word_length = max_word_length
        self.lower = lowercase
        self.vocabs = {'word': None,
                       'char': None,
                       'pos': None,
                       'chunk': None}
        self._data_dict = {}

    def _validate_paths(self, data_path):
        validate_existing_directory(data_path)
        for f in self.dataset_files:
            _f_path = path.join(data_path, self.dataset_files[f])
            validate_existing_filepath(_f_path)
            self.dataset_files[f] = _f_path

    def _load_data(self):
        """
        open files and parse
        return format: list of 3-tuples (word list, POS list, chunk list)
        """
        train_set = read_sequential_tagging_file(self.dataset_files['train'])
        test_set = read_sequential_tagging_file(self.dataset_files['test'])
        train_data = [list(zip(*x)) for x in train_set]
        test_data = [list(zip(*x)) for x in test_set]
        return train_data, test_data

    @property
    def train_set(self):
        """get the train set"""
        if self._data_dict.get('train', None) is None:
            self._gen_data()
        return self._data_dict.get('train')

    @property
    def test_set(self):
        """get the test set"""
        if self._data_dict.get('test', None) is None:
            self._gen_data()
        return self._data_dict.get('test')

    @staticmethod
    def _extract(x, y, n):
        return list(zip(*x))[n] + list(zip(*y))[n]

    @property
    def word_vocab(self):
        """word Vocabulary"""
        return self.vocabs['word']

    @property
    def char_vocab(self):
        """character Vocabulary"""
        return self.vocabs['char']

    @property
    def pos_vocab(self):
        """pos label Vocabulary"""
        return self.vocabs['pos']

    @property
    def chunk_vocab(self):
        """chunk label Vocabulary"""
        return self.vocabs['chunk']

    def _gen_data(self):
        train, test = self._load_data()
        train_size = len(train)
        test_size = len(test)
        sentences = self._extract(train, test, 0)
        pos_tags = self._extract(train, test, 1)
        chunk_tags = self._extract(train, test, 2)
        sentence_vecs, word_vocab = word_vector_generator(sentences, self.lower, 2)
        pos_vecs, pos_vocab = word_vector_generator(pos_tags, start=1)
        chunk_vecs, chunk_vocab = word_vector_generator(chunk_tags, start=1)
        self.vocabs = {'word': word_vocab,  # 0=pad, 1=unk
                       'pos': pos_vocab,  # 0=pad, 1=unk
                       'chunk': chunk_vocab}  # 0=pad
        if self.sentence_length is not None:
            sentence_vecs = pad_sentences(sentence_vecs, max_length=self.sentence_length)
            chunk_vecs = pad_sentences(chunk_vecs, max_length=self.sentence_length)
            pos_vecs = pad_sentences(pos_vecs, max_length=self.sentence_length)
        self._data_dict['train'] = sentence_vecs[:train_size], pos_vecs[:train_size], \
            chunk_vecs[:train_size]
        self._data_dict['test'] = sentence_vecs[-test_size:], pos_vecs[-test_size:], \
            chunk_vecs[-test_size:]
        if self.use_chars:
            chars_vecs, char_vocab = character_vector_generator(sentences, start=2)
            self.vocabs.update({'char': char_vocab})  # 0=pad, 1=unk
            if self.max_word_length is not None:
                chars_vecs = [pad_sentences(d, max_length=self.max_word_length)
                              for d in chars_vecs]
                zeros = np.zeros((len(chars_vecs), self.sentence_length, self.max_word_length))
                for idx, d in enumerate(chars_vecs):
                    d = d[:self.sentence_length]
                    zeros[idx, -d.shape[0]:] = d
                chars_vecs = zeros.astype(dtype=np.int32)
            self._data_dict['train'] += (chars_vecs[:train_size],)
            self._data_dict['test'] += (chars_vecs[-test_size:],)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from nlp_architect.api.abstract_api import AbstractApi
from nlp_architect.pipelines.spacy_bist import SpacyBISTParser


class BistParserApi(AbstractApi):
    """
    Bist Parser API
    """
    def __init__(self):
        self.model = None

    def load_model(self):
        """
        Load SpacyBISTParser model
        """
        self.model = SpacyBISTParser()

    def inference(self, doc):
        """
        Parse according to SpacyBISTParser's model

        Args:
            doc (str): the doc str

        Returns:
            CoreNLPDoc: the parser's response hosted in CoreNLPDoc object
        """
        return self.model.parse(doc)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

import numpy as np
import re


# pylint: disable=invalid-unary-operand-type
def pad_sentences(sequences: np.ndarray,
                  max_length: int = None,
                  padding_value: int = 0,
                  padding_style='post') -> np.ndarray:
    """
    Pad input sequences up to max_length
    values are aligned to the right

    Args:
        sequences (iter): a 2D matrix (np.array) to pad
        max_length (int, optional): max length of resulting sequences
        padding_value (int, optional): padding value
        padding_style (str, optional): add padding values as prefix (use with 'pre')
            or postfix (use with 'post')

    Returns:
        input sequences padded to size 'max_length'
    """
    if isinstance(sequences, list) and len(sequences) > 0:
        try:
            sequences = np.asarray(sequences)
        except ValueError:
            print('cannot convert sequences into numpy array')
    assert hasattr(sequences, 'shape')
    if len(sequences) < 1:
        return sequences
    if max_length is None:
        max_length = np.max([len(s) for s in sequences])
    elif max_length < 1:
        raise ValueError('max sequence length must be > 0')
    if max_length < 1:
        return sequences
    padded_sequences = (np.ones((len(sequences), max_length), dtype=np.int32) * padding_value)
    for i, sent in enumerate(sequences):
        if padding_style == 'post':
            trunc = sent[-max_length:]
            padded_sequences[i, :len(trunc)] = trunc
        elif padding_style == 'pre':
            trunc = sent[:max_length]
            padded_sequences[i, -trunc:] = trunc
    return padded_sequences.astype(dtype=np.int32)


def one_hot(mat: np.ndarray, num_classes: int) -> np.ndarray:
    """
    Convert a 1D matrix of ints into one-hot encoded vectors.

    Arguments:
        mat (numpy.ndarray): A 1D matrix of labels (int)
        num_classes (int): Number of all possible classes

    Returns:
        numpy.ndarray: A 2D matrix
    """
    assert len(mat.shape) < 2 or isinstance(mat.shape, int)
    vec = np.zeros((mat.shape[0], num_classes))
    for i, v in enumerate(mat):
        vec[i][v] = 1.0
    return vec


def one_hot_sentence(mat: np.ndarray, num_classes: int) -> np.ndarray:
    """
    Convert a 2D matrix of ints into one-hot encoded 3D matrix

    Arguments:
        mat (numpy.ndarray): A 2D matrix of labels (int)
        num_classes (int): Number of all possible classes

    Returns:
        numpy.ndarray: A 3D matrix
    """
    new_mat = []
    for i in range(mat.shape[0]):
        new_mat.append(one_hot(mat[i], num_classes))
    return np.asarray(new_mat)


def add_offset(mat: np.ndarray, offset: int = 1) -> np.ndarray:
    """
    Add +1 to all values in matrix mat

    Arguments:
        mat (numpy.ndarray): A 2D matrix with int values
        offset (int): offset to add

    Returns:
        numpy.ndarray: input matrix
    """
    for i, vec in enumerate(mat):
        offset_arr = np.array(vec.shape)
        offset_arr.fill(offset)
        mat[i] = vec + offset_arr
    return mat


def license_prompt(model_name, model_website, dataset_dir=None):
    if dataset_dir:
        print('\n\n***\n{} was not found in the directory: {}'.format(model_name, dataset_dir))
    else:
        print('\n\n***\n\n{} was not found on local installation'.format(model_name))
    print('{} can be downloaded from {}'.format(model_name, model_website))
    print('The terms and conditions of the data set license apply. Intel does not '
          'grant any rights to the data files or database\n')
    response = input('To download \'{}\' from {}, please enter YES: '.
                     format(model_name, model_website))
    res = response.lower().strip()
    if res == "yes" or (len(res) == 1 and res == 'y'):
        print('Downloading {}...'.format(model_name))
        responded_yes = True
    else:
        print('Download declined. Response received {} != YES|Y. '.format(res))
        if dataset_dir:
            print('Please download the model manually from {} and place in the directory: {}'
                  .format(model_website, dataset_dir))
        else:
            print('Please download the model manually from {}'.format(model_website))
        responded_yes = False
    return responded_yes


# character vocab
zhang_lecun_vocab = list("abcdefghijklmnopqrstuvwxyz0123456789-,;.!?:’/\\|_@#$%ˆ&*˜‘+=<>()[]{}")
vocab_hash = {b: a for a, b in enumerate(zhang_lecun_vocab)}


def normalize(txt, vocab=None, replace_char=' ',
              max_length=300, pad_out=True,
              to_lower=True, reverse=False,
              truncate_left=False, encoding=None):

    # remove html
    # This will keep characters and other symbols
    txt = txt.split()
    # Remove HTML
    txt = [re.sub(r'http:.*', '', r) for r in txt]
    txt = [re.sub(r'https:.*', '', r) for r in txt]

    txt = (" ".join(txt))

    # Remove punctuation
    txt = re.sub("[.,!]", " ", txt)
    txt = " ".join(txt.split())

    # store length for multiple comparisons
    txt_len = len(txt)

    if truncate_left:
        txt = txt[-max_length:]
    else:
        txt = txt[:max_length]
    # change case
    if to_lower:
        txt = txt.lower()
    # Reverse order
    if reverse:
        txt = txt[::-1]
    # replace chars
    if vocab is not None:
        txt = ''.join([c if c in vocab else replace_char for c in txt])
    # re-encode text
    if encoding is not None:
        txt = txt.encode(encoding, errors="ignore")
    # pad out if needed
    if pad_out and max_length > txt_len:
        txt = txt + replace_char * (max_length - txt_len)
    return txt


def to_one_hot(txt, vocab=vocab_hash):
    vocab_size = len(vocab.keys())
    one_hot_vec = np.zeros((vocab_size + 1, len(txt)), dtype=np.float32)
    # run through txt and "switch on" relevant positions in one-hot vector
    for idx, char in enumerate(txt):
        if char in vocab_hash:
            vocab_idx = vocab_hash[char]
            one_hot_vec[vocab_idx, idx] = 1
        # raised if character is out of vocabulary
        else:
            pass
    return one_hot_vec


def balance(df):
    print("Balancing the classes")
    type_counts = df['Sentiment'].value_counts()
    min_count = min(type_counts.values)

    balanced_df = None
    for key in type_counts.keys():
        df_sub = df[df['Sentiment'] == key].sample(n=min_count, replace=False)
        if balanced_df is not None:
            balanced_df = balanced_df.append(df_sub)
        else:
            balanced_df = df_sub
    return balanced_df



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from os import path, remove, makedirs

from nlp_architect.common.core_nlp_doc import CoreNLPDoc
from nlp_architect.data.conll import ConllEntry
from nlp_architect.models.bist_parser import BISTModel
from nlp_architect import LIBRARY_OUT
from nlp_architect.utils.io import download_unlicensed_file, uncompress_file
from nlp_architect.utils.io import validate
from nlp_architect.utils.text import SpacyInstance


class SpacyBISTParser(object):
    """Main class which handles parsing with Spacy-BIST parser.

    Args:
        verbose (bool, optional): Controls output verbosity.
        spacy_model (str, optional): Spacy model to use
        (see https://spacy.io/api/top-level#spacy.load).
        bist_model (str, optional): Path to a .model file to load. Defaults pre-trained model'.
    """
    dir = LIBRARY_OUT / 'bist-pretrained'
    _pretrained = dir / 'bist.model'

    def __init__(self, verbose=False, spacy_model='en', bist_model=None):
        validate((verbose, bool), (spacy_model, str, 0, 1000),
                 (bist_model, (type(None), str), 0, 1000))
        if not bist_model:
            print("Using pre-trained BIST model.")
            _download_pretrained_model()
            bist_model = SpacyBISTParser._pretrained

        self.verbose = verbose
        self.bist_parser = BISTModel()
        self.bist_parser.load(bist_model if bist_model else SpacyBISTParser._pretrained)
        self.spacy_parser = SpacyInstance(spacy_model,
                                          disable=['ner', 'vectors', 'textcat']).parser

    def to_conll(self, doc_text):
        """Converts a document to CoNLL format with spacy POS tags.

        Args:
            doc_text (str): raw document text.

        Yields:
            list of ConllEntry: The next sentence in the document in CoNLL format.
        """
        validate((doc_text, str))
        for sentence in self.spacy_parser(doc_text).sents:
            sentence_conll = [ConllEntry(0, '*root*', '*root*', 'ROOT-POS', 'ROOT-CPOS', '_',
                                         -1, 'rroot', '_', '_')]
            i_tok = 0
            for tok in sentence:
                if self.verbose:
                    print(tok.text + '\t' + tok.tag_)

                if not tok.is_space:
                    pos = tok.tag_
                    text = tok.text

                    if text != '-' or pos != 'HYPH':
                        pos = _spacy_pos_to_ptb(pos, text)
                        token_conll = ConllEntry(i_tok + 1, text, tok.lemma_, pos, pos,
                                                 tok.ent_type_, -1, '_', '_', tok.idx)
                        sentence_conll.append(token_conll)
                        i_tok += 1

            if self.verbose:
                print('-----------------------\ninput conll form:')
                for entry in sentence_conll:
                    print(str(entry.id) + '\t' + entry.form + '\t' + entry.pos + '\t')
            yield sentence_conll

    def parse(self, doc_text, show_tok=True, show_doc=True):
        """Parse a raw text document.

        Args:
            doc_text (str)
            show_tok (bool, optional): Specifies whether to include token text in output.
            show_doc (bool, optional): Specifies whether to include document text in output.

        Returns:
            CoreNLPDoc: The annotated document.
        """
        validate((doc_text, str), (show_tok, bool), (show_doc, bool))
        doc_conll = self.to_conll(doc_text)
        parsed_doc = CoreNLPDoc()

        if show_doc:
            parsed_doc.doc_text = doc_text

        for sent_conll in self.bist_parser.predict_conll(doc_conll):
            parsed_sent = []
            conj_governors = {'and': set(), 'or': set()}

            for tok in sent_conll:
                gov_id = int(tok.pred_parent_id)
                rel = tok.pred_relation

                if tok.form != '*root*':
                    if tok.form.lower() == 'and':
                        conj_governors['and'].add(gov_id)
                    if tok.form.lower() == 'or':
                        conj_governors['or'].add(gov_id)

                    if rel == 'conj':
                        if gov_id in conj_governors['and']:
                            rel += '_and'
                        if gov_id in conj_governors['or']:
                            rel += '_or'

                    parsed_tok = {'start': tok.misc, 'len': len(tok.form),
                                  'pos': tok.pos, 'ner': tok.feats,
                                  'lemma': tok.lemma, 'gov': gov_id - 1,
                                  'rel': rel}

                    if show_tok:
                        parsed_tok['text'] = tok.form
                    parsed_sent.append(parsed_tok)
            if parsed_sent:
                parsed_doc.sentences.append(parsed_sent)
        return parsed_doc


def _download_pretrained_model():
    """Downloads the pre-trained BIST model if non-existent."""
    if not path.isfile(SpacyBISTParser.dir / 'bist.model'):
        print('Downloading pre-trained BIST model...')
        zip_path = SpacyBISTParser.dir / 'bist-pretrained.zip'
        makedirs(SpacyBISTParser.dir, exist_ok=True)
        download_unlicensed_file(
            'https://s3-us-west-2.amazonaws.com/nlp-architect-data/models/dep_parse/',
            'bist-pretrained.zip', zip_path)
        print('Unzipping...')
        uncompress_file(zip_path, outpath=str(SpacyBISTParser.dir))
        remove(zip_path)
        print('Done.')


def _spacy_pos_to_ptb(pos, text):
    """
    Converts a Spacy part-of-speech tag to a Penn Treebank part-of-speech tag.

    Args:
        pos (str): Spacy POS tag.
        text (str): The token text.

    Returns:
        ptb_tag (str): Standard PTB POS tag.
    """
    validate((pos, str, 0, 30), (text, str, 0, 1000))
    ptb_tag = pos
    if text in ['...', '—']:
        ptb_tag = ':'
    elif text == '*':
        ptb_tag = 'SYM'
    elif pos == 'AFX':
        ptb_tag = 'JJ'
    elif pos == 'ADD':
        ptb_tag = 'NN'
    elif text != pos and text in [',', '.', ":", '``', '-RRB-', '-LRB-']:
        ptb_tag = text
    elif pos in ['NFP', 'HYPH', 'XX']:
        ptb_tag = 'SYM'
    return ptb_tag



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import logging
import os
from typing import List, Set

from nlp_architect.common.cdc.mention_data import MentionData
from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.data.cdc_resources.relations.relation_types_enums import RelationType
from nlp_architect.utils.io import load_json_file

logger = logging.getLogger(__name__)


class WithinDocCoref(RelationExtraction):
    def __init__(self, wd_file: str):
        """
        Extract Relation between two mentions according to Within document co-reference

        Args:
            wd_file (required): str Location of within doc co-reference mentions file
        """
        logger.info('Loading Within doc resource')
        if wd_file is not None and os.path.isfile(wd_file):
            wd_mentions_json = load_json_file(wd_file)
            self.within_doc_coref_chain = self.arrange_resource(wd_mentions_json)
        else:
            raise FileNotFoundError('Within-doc resource file not found or not in path')
        super(WithinDocCoref, self).__init__()

    @staticmethod
    def arrange_resource(wd_mentions_json):
        document_tokens_dict = dict()
        for mention_json in wd_mentions_json:
            mention_data = MentionData.read_json_mention_data_line(mention_json)
            mention_tokens = mention_data.tokens_number
            for i in range(0, len(mention_tokens)):
                doc_id = mention_data.doc_id
                sent_id = mention_data.sent_id
                token_map_key = MentionData.static_gen_token_unique_id(doc_id, sent_id,
                                                                       mention_tokens[i])
                document_tokens_dict[token_map_key] = mention_data.coref_chain
        return document_tokens_dict

    def extract_all_relations(self, mention_x: MentionData,
                              mention_y: MentionData) -> Set[RelationType]:
        ret_ = set()
        ret_.add(self.extract_sub_relations(mention_x, mention_y, RelationType.WITHIN_DOC_COREF))
        return ret_

    def extract_sub_relations(self, mention_x: MentionData, mention_y: MentionData,
                              relation: RelationType) -> RelationType:
        """
        Check if input mentions has the given relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight
            relation: RelationType

        Returns:
            RelationType: relation in case mentions has given relation or
                RelationType.NO_RELATION_FOUND otherwise
        """
        if relation is not RelationType.WITHIN_DOC_COREF:
            return RelationType.NO_RELATION_FOUND

        if mention_x.doc_id == mention_y.doc_id:
            ment_x_coref_chain = self.extract_within_coref(mention_x)
            ment_y_coref_chain = self.extract_within_coref(mention_y)

            if not ment_x_coref_chain or not ment_y_coref_chain:
                return RelationType.NO_RELATION_FOUND

            if '-' in ment_x_coref_chain or '-' in ment_y_coref_chain:
                return RelationType.NO_RELATION_FOUND

            if set(ment_x_coref_chain) == set(ment_y_coref_chain):
                return RelationType.WITHIN_DOC_COREF

        return RelationType.NO_RELATION_FOUND

    def extract_within_coref(self, mention: MentionData) -> List[str]:
        tokens = mention.tokens_number
        within_coref_token = []
        for token_id in tokens:
            token_x_id = MentionData.static_gen_token_unique_id(str(mention.doc_id),
                                                                str(mention.sent_id),
                                                                str(token_id))
            if token_x_id in self.within_doc_coref_chain:
                token_coref_chain = self.within_doc_coref_chain[token_x_id]
                if token_coref_chain:
                    within_coref_token.append(token_coref_chain)
            else:
                within_coref_token.append('-')
                break

        return within_coref_token

    def get_within_doc_coref_chain(self):
        return self.within_doc_coref_chain

    @staticmethod
    def create_ment_id(mention_x: MentionData, mention_y: MentionData) -> str:
        return '_'.join([mention_x.get_mention_id(), mention_y.get_mention_id()])

    @staticmethod
    def get_supported_relations() -> List[RelationType]:
        """
        Return all supported relations by this class

        Returns:
            List[RelationType]
        """
        return [RelationType.WITHIN_DOC_COREF]



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import os
from typing import List

from nlp_architect.common.cdc.cluster import Clusters
from nlp_architect.common.cdc.mention_data import MentionData
from nlp_architect.common.cdc.topics import Topic
from nlp_architect.utils.string_utils import StringUtils

logger = logging.getLogger(__name__)


def write_clusters_to_file(clusters: Clusters, topic_id: str, file_obj) -> None:
    """
    Write the clusters to a text file (for experiments or evaluation using
        coreference scorer (v8.01))
    Args:
        clusters: the cluster to write
        topic_id:
        file_obj: file object
    """
    i = 0
    file_obj.write('Topic - ' + topic_id + '\n')
    for cluster in clusters.clusters_list:
        i += 1
        file_obj.write('cluster #' + str(i) + '\n')
        mentions_list = []
        for mention in cluster.mentions:
            mentions_list.append((mention.tokens_str, mention.predicted_coref_chain))
        file_obj.write(str(mentions_list) + '\n')


def extract_vocab(mentions: List[MentionData], filter_stop_words: bool) -> List[str]:
    """
    Extract Head, Lemma and mention string from all mentions to create a list of string vocabulary
    Args:
        mentions:
        filter_stop_words:

    Returns:

    """
    vocab = set()
    for mention in mentions:
        head = mention.mention_head
        head_lemma = mention.mention_head_lemma
        tokens_str = mention.tokens_str
        if not filter_stop_words:
            vocab.add(head)
            vocab.add(head_lemma)
            vocab.add(tokens_str)
        else:
            if not StringUtils.is_stop(head):
                vocab.add(head)
            if not StringUtils.is_stop(head_lemma):
                vocab.add(head_lemma)
            if not StringUtils.is_stop(tokens_str):
                vocab.add(tokens_str)
    vocab_set = list(vocab)
    return vocab_set


def load_mentions_vocab_from_files(mentions_files, filter_stop_words=False):
    logger.info('Loading mentions files...')
    mentions = []
    for _file in mentions_files:
        mentions.extend(MentionData.read_mentions_json_to_mentions_data_list(_file))

    return load_mentions_vocab(mentions, filter_stop_words)


def load_mentions_vocab(mentions, filter_stop_words=False):
    vocab = extract_vocab(mentions, filter_stop_words)
    logger.info('Done loading mentions files...')
    return vocab


def write_event_coref_scorer_results(topics_list: List[Topic], output_file: str) -> None:
    with open(os.path.join(output_file, 'cd_event_pred_clusters_spans.txt'), 'w') as output:
        write_topics(topics_list, output)


def write_entity_coref_scorer_results(topics_list: List[Topic], output_file: str) -> None:
    with open(os.path.join(output_file, 'cd_entity_pred_clusters_spans.txt'), 'w') as output:
        write_topics(topics_list, output)


def write_topics(topics_list: List[Topic], output) -> None:
    output.write('#begin document (ECB+/ecbplus_all); part 000\n')
    for topic in topics_list:
        for mention in topic.mentions:
            output.write('ECB+/ecbplus_all\t' + '(' + str(mention.predicted_coref_chain) + ')\n')
    output.write('#end document')



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import math
from os import PathLike

from nlp_architect.common.core_nlp_doc import CoreNLPDoc
from nlp_architect.models.absa import INFERENCE_OUT
from nlp_architect.models.absa.inference.data_types import Term, TermType, Polarity, SentimentDoc,\
    SentimentSentence, LexiconElement
from nlp_architect.models.absa.utils import _read_lexicon_from_csv, load_opinion_lex, \
    _load_aspect_lexicon

INTENSIFIER_FACTOR = 0.3
VERB_POS = {"VB", "VBD", "VBG", "VBN", "VBP", "VBZ"}


class SentimentInference(object):
    """Main class for sentiment inference execution.

    Attributes:
        opinion_lex: Opinion lexicon as outputted by TrainSentiment module.
        aspect_lex: Aspect lexicon as outputted by TrainSentiment module.
        intensifier_lex (dict): Pre-defined intensifier lexicon.
        negation_lex (dict): Pre-defined negation lexicon.
    """

    def __init__(self, aspect_lex: PathLike, opinion_lex: PathLike or dict, parse: bool = True):
        """Inits SentimentInference with given aspect and opinion lexicons."""
        INFERENCE_OUT.mkdir(parents=True, exist_ok=True)
        self.opinion_lex = \
            opinion_lex if type(opinion_lex) is dict else load_opinion_lex(opinion_lex)
        self.aspect_lex = _load_aspect_lexicon(aspect_lex)
        self.intensifier_lex = _read_lexicon_from_csv('IntensifiersLex.csv')
        self.negation_lex = _read_lexicon_from_csv('NegationSentLex.csv')

        if parse:
            from nlp_architect.pipelines.spacy_bist import SpacyBISTParser
            self.parser = SpacyBISTParser()
        else:
            self.parser = None

    def run(self, doc: str = None, parsed_doc: CoreNLPDoc = None) -> SentimentDoc:
        """Run SentimentInference on a single document.

        Returns:
            The sentiment annotated document, which contains the detected events per sentence.
        """
        if not parsed_doc:
            if not self.parser:
                raise RuntimeError("Parser not initialized (try parse=True at init )")
            parsed_doc = self.parser.parse(doc)

        sentiment_doc = None
        for sentence in parsed_doc.sentences:
            events = []
            scores = []
            for aspect_row in self.aspect_lex:
                _, asp_events = self._extract_event(aspect_row, sentence)
                for asp_event in asp_events:
                    events.append(asp_event)
                    scores += [term.score for term in asp_event if term.type == TermType.ASPECT]

            if events:
                if not sentiment_doc:
                    sentiment_doc = SentimentDoc(parsed_doc.doc_text)
                sentiment_doc.sentences.append(
                    SentimentSentence(sentence[0]['start'],
                                      sentence[-1]['start'] + sentence[-1]['len'] - 1, events))
        return sentiment_doc

    def _extract_intensifier_terms(self, toks, sentiment_index, polarity, sentence):
        """Extract intensifier events from sentence."""
        count = 0
        terms = []
        for intens_i, intens in [(i, x) for i, x in enumerate(toks) if x in self.intensifier_lex]:
            if math.fabs(sentiment_index - intens_i) == 1:
                score = self.intensifier_lex[intens].score
                terms.append(Term(intens, TermType.INTENSIFIER, polarity, score,
                                  sentence[intens_i]['start'], sentence[intens_i]['len']))
                count += abs(score + float(INTENSIFIER_FACTOR))
        return count if count != 0 else 1, terms

    def _extract_neg_terms(self, toks: list, op_i: int, sentence: list) -> tuple:
        """Extract negation terms from sentence.

        Args:
            toks: Sentence text broken down to tokens (words).
            op_i: Index of opinion term in sentence.
            sentence: parsed sentence

        Returns:
            List of negation terms and its aggregated sign (positive or negative).
        """
        sign = 1
        terms = []
        gov_op_i = sentence[op_i]['gov']
        dep_op_indices = [sentence.index(x) for x in sentence if x['gov'] == op_i]
        for neg_i, negation in [(i, x) for i, x in enumerate(toks) if x in self.negation_lex]:
            position = self.negation_lex[negation].position
            dist = op_i - neg_i
            before = position == 'before' and (dist == 1 or neg_i in dep_op_indices)
            after = position == 'after' and (dist == -1 or neg_i == gov_op_i)
            both = position == 'both' and dist in (1, -1)
            if before or after or both:
                terms.append(Term(negation, TermType.NEGATION, Polarity.NEG,
                                  self.negation_lex[negation].score,
                                  sentence[toks.index(negation)]['start'],
                                  sentence[toks.index(negation)]['len']))
                sign *= self.negation_lex[negation].score
        return terms, sign

    def _extract_event(self, aspect_row: LexiconElement, parsed_sentence: list) -> tuple:
        """Extract opinion and aspect terms from sentence."""
        event = []
        sent_aspect_pair = None
        real_aspect_indices = _consolidate_aspects(aspect_row.term, parsed_sentence)
        aspect_key = aspect_row.term[0]
        for aspect_index_range in real_aspect_indices:
            for word_index in aspect_index_range:
                sent_aspect_pair, event = \
                    self._detect_opinion_aspect_events(word_index, parsed_sentence, aspect_key,
                                                       aspect_index_range)
                if sent_aspect_pair:
                    break
        return sent_aspect_pair, event

    @staticmethod
    def _modify_for_multiple_word(cur_tkn, parsed_sentence, index_range):
        """Modify multiple-word aspect tkn length and start index.

        Args:
            index_range: The index range of the multi-word aspect.
        Returns:
            The modified aspect token.
        """
        if len(index_range) >= 2:
            cur_tkn["start"] = parsed_sentence[index_range[0]]["start"]
            cur_tkn["len"] = len(parsed_sentence[index_range[0]]["text"])
            for i in index_range[1:]:
                cur_tkn["len"] = int(cur_tkn["len"]) + len(
                    parsed_sentence[i]["text"]) + 1
        return cur_tkn

    def _detect_opinion_aspect_events(self, aspect_index, parsed_sent, aspect_key, index_range):
        """Extract opinion-aspect events from sentence.

        Args:
            aspect_index: index of aspect in sentence.
            parsed_sent: current sentence parse tree.
            aspect_key: main aspect term serves as key in aspect dict.
            index_range: The index range of the multi word aspect.

        Returns:
            List of aspect sentiment pair, and list of events extracted.
        """
        all_pairs, events = [], []
        sentence_text_list = [x["text"] for x in parsed_sent]
        sentence_text = ' '.join(sentence_text_list)
        cur_aspect = parsed_sent[aspect_index]["text"]
        for tok in parsed_sent:
            aspect_op_pair = []
            terms = []
            pos = tok['pos']
            gov_i = tok['gov']
            gov = parsed_sent[gov_i]
            gov_text = gov['text']
            tok_text = tok['text']

            # is cur_tkn an aspect and gov a opinion?
            if tok_text.lower() == cur_aspect.lower() and parsed_sent.index(tok) == aspect_index \
                    and gov_text.lower() in self.opinion_lex and \
                    gov['pos'] not in VERB_POS:
                aspect_op_pair.append(
                    (self._modify_for_multiple_word(tok, parsed_sent, index_range), gov))
            #  is gov an aspect and cur_tkn a opinion?
            if gov_text.lower() == cur_aspect.lower() and gov_i == aspect_index \
                    and tok_text.lower() in self.opinion_lex and pos not in VERB_POS:
                aspect_op_pair.append(
                    (self._modify_for_multiple_word(gov, parsed_sent, index_range), tok))
            # if aspect_tok found
            for aspect, opinion in aspect_op_pair:
                op_tok_i = parsed_sent.index(opinion)
                score = self.opinion_lex[opinion['text'].lower()].score
                neg_terms, sign = self._extract_neg_terms(sentence_text_list, op_tok_i,
                                                          parsed_sent)
                polarity = Polarity.POS if score * sign > 0 else Polarity.NEG
                intensifier_score, intensifier_terms = self._extract_intensifier_terms(
                    sentence_text_list, op_tok_i, polarity, parsed_sent)
                over_all_score = score * sign * intensifier_score
                terms.append(Term(aspect_key, TermType.ASPECT, polarity, over_all_score,
                                  aspect['start'], aspect['len']))
                terms.append(Term(opinion['text'], TermType.OPINION, polarity, over_all_score,
                                  opinion['start'], opinion['len']))
                if len(neg_terms) > 0:
                    terms = terms + neg_terms
                if len(intensifier_terms) > 0:
                    terms = terms + intensifier_terms
                all_pairs.append([aspect_key, opinion['text'], over_all_score, polarity,
                                  sentence_text])
                events.append(terms)
        return all_pairs, events


def _sentence_contains_after(sentence, index, phrase):
    """Returns sentence contains phrase after given index."""
    for i in range(len(phrase)):
        if len(sentence) <= index + i or len(phrase) <= i or \
                sentence[index + i]['text'].lower() != phrase[i].lower():
            return False
    return True


def _consolidate_aspects(aspect_row, sentence):
    """Returns consolidated indices of aspect terms in sentence.

    Args:
        aspect_row: List of aspect terms which belong to the same aspect-group.
    """
    indices = []
    aspect_phrases: list = \
        sorted([phrase.split(' ') for phrase in aspect_row], key=len, reverse=True)
    appeared = set()
    for tok_i in range(len(sentence)):
        for aspect_phrase in aspect_phrases:
            if _sentence_contains_after(sentence, tok_i, aspect_phrase):
                span = range(tok_i, tok_i + len(aspect_phrase))
                if not appeared & set(span):
                    appeared |= set(span)
                    indices.append(list(span))
    return indices



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
from os import listdir
from os.path import join, isfile

from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_page import WikipediaPage
from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_page_extracted_relations import \
    WikipediaPageExtractedRelations
from nlp_architect.data.cdc_resources.wikipedia.wiki_search_page_result import \
    WikipediaSearchPageResult
from nlp_architect.utils.io import load_json_file

logger = logging.getLogger(__name__)


class WikiOffline(object):
    def __init__(self, wikidump):
        if wikidump:
            self.dump = self.load_dump(wikidump)
            logger.info('Wikipedia dump loaded successfully!')

    def get_pages(self, phrase):
        if phrase and phrase in self.dump:
            pages = self.dump[phrase]
            if pages:
                return pages

        return set()

    @staticmethod
    def extract_json_values(json_pages):
        pages = set()
        for json_page in json_pages:
            description = json_page.get('description', None)
            pageid = int(json_page.get('pageid', 0))
            orig_phrase = json_page.get('orig_phrase', None)
            orig_phrase_norm = json_page.get('orig_phrase_norm', None)
            wiki_title = json_page.get('wiki_title', None)
            wiki_title_norm = json_page.get('wiki_title_norm', None)

            relations_json = json_page.get('relations', None)
            rel_is_part_name = relations_json.get('isPartName', None)
            rel_is_disambiguation = relations_json.get('isDisambiguation', None)
            rel_disambiguation = relations_json.get('disambiguationLinks', None)
            rel_disambiguation_norm = relations_json.get('disambiguationLinksNorm', None)
            rel_parenthesis = relations_json.get('titleParenthesis', None)
            rel_parenthesis_norm = relations_json.get('titleParenthesisNorm', None)
            rel_categories = relations_json.get('categories', None)
            rel_categories_norm = relations_json.get('categoriesNorm', None)
            rel_be_comp = relations_json.get('beCompRelations', None)
            rel_be_comp_norm = relations_json.get('beCompRelationsNorm', None)
            rel_aliases = relations_json.get('aliases', None)
            rel_aliases_norm = relations_json.get('aliasesNorm', None)

            relations = WikipediaPageExtractedRelations(rel_is_part_name, rel_is_disambiguation,
                                                        rel_parenthesis,
                                                        rel_disambiguation,
                                                        rel_categories, rel_aliases, rel_be_comp,
                                                        rel_disambiguation_norm,
                                                        rel_categories_norm, rel_aliases_norm,
                                                        rel_parenthesis_norm,
                                                        rel_be_comp_norm)

            page = WikipediaPage(orig_phrase, orig_phrase_norm, wiki_title, wiki_title_norm, 0,
                                 pageid, description,
                                 relations)
            pages.add(WikipediaSearchPageResult(orig_phrase, page))

        return pages

    def load_dump(self, wiki_dump):
        onlyfiles = []
        for _file in listdir(wiki_dump):
            file_path = join(wiki_dump, _file)
            if isfile(file_path):
                onlyfiles.append(file_path)

        json_dump_list = {}
        for _file in onlyfiles:
            json_dump_list.update(load_json_file(_file))

        dump_final = {}
        for key, value in json_dump_list.items():
            dump_final[key] = self.extract_json_values(value)

        return dump_final

    class NoPage(object):
        """ Attribute not found. """

        def __init__(self, *args, **kwargs):  # real signature unknown
            pass

        @staticmethod  # known case of __new__
        def __new__(S, *more):  # real signature unknown; restored from __doc__
            """ T.__new__(S, ...) -> a new object with type S, a subtype of T """



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import pickle

import numpy as np

from nlp_architect.common.cdc.mention_data import MentionDataLight

logger = logging.getLogger(__name__)


class GloveEmbedding(object):
    def __init__(self, glove_file):
        logger.info('Loading Glove Online Embedding module, This my take a while...')
        self.word_to_ix, self.word_embeddings = self.load_glove_for_vocab(glove_file)
        logger.info('Glove Offline Embedding module lead successfully')

    @staticmethod
    def load_glove_for_vocab(glove_filename):
        vocab = []
        embd = []
        with open(glove_filename) as glove_file:
            for line in glove_file:
                row = line.strip().split(' ')
                word = row[0]
                vocab.append(word)
                embd.append(row[1:])

        embeddings = np.asarray(embd, dtype=float)
        word_to_ix = {word: i for i, word in enumerate(vocab)}
        return word_to_ix, embeddings


class GloveEmbeddingOffline(object):
    def __init__(self, embed_resources):
        logger.info('Loading Glove Offline Embedding module')
        with open(embed_resources, 'rb') as out:
            self.word_to_ix, self.word_embeddings = pickle.load(out, encoding='latin1')
        logger.info('Glove Offline Embedding module lead successfully')

    def get_feature_vector(self, mention: MentionDataLight):
        embed = None
        head = mention.mention_head
        lemma = mention.mention_head_lemma
        if head in self.word_to_ix:
            embed = self.word_embeddings[self.word_to_ix[head]]
        elif lemma in self.word_to_ix:
            embed = self.word_embeddings[self.word_to_ix[lemma]]

        return embed

    def get_avrg_feature_vector(self, tokens_str):
        embed = np.zeros(300, dtype=np.float64)
        mention_size = 0
        for token in tokens_str.split():
            if token in self.word_to_ix:
                token_embed = self.word_embeddings[self.word_to_ix[token]]
                embed = np.add(embed, token_embed)
                mention_size += 1

        if mention_size == 0:
            mention_size = 1

        return np.true_divide(embed, mention_size)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import division
from __future__ import print_function

import os
import re
import zipfile
from os import makedirs
from random import shuffle

import numpy as np
import tensorflow as tf

from nlp_architect.api.abstract_api import AbstractApi
from nlp_architect.models.matchlstm_ansptr import MatchLSTMAnswerPointer
from nlp_architect import LIBRARY_OUT
from nlp_architect.utils.generic import license_prompt
from nlp_architect.utils.io import download_unlicensed_file
from nlp_architect.utils.mrc_utils import (
    create_squad_training, max_values_squad, get_data_array_squad)


class MachineComprehensionApi(AbstractApi):
    """
    Machine Comprehension API
    """
    dir = str(LIBRARY_OUT / 'mrc-pretrained')
    data_path = os.path.join(dir, 'mrc_data', 'data')
    data_dir = os.path.join(dir, 'mrc_data')
    model_dir = os.path.join(dir, 'mrc_trained_model')
    model_path = os.path.join(dir, 'mrc_trained_model', 'trained_model')

    def __init__(self, prompt=True):
        self.prompt = None
        self.vocab_dict = None
        self.vocab_rev = None
        self.model = None
        self.dev = None
        self.sess = None
        self.prompt = prompt
        self.params_dict = {'batch_size': 1,
                            'hidden_size': 150,
                            'max_para': 300,
                            'epoch_no': 15,
                            'inference_only': True}
        self.file_name_dict = {'train_para_ids': 'train.ids.context',
                               'train_ques_ids': 'train.ids.question',
                               'train_answer': 'train.span',
                               'val_para_ids': 'dev.ids.context',
                               'val_ques_ids': 'dev.ids.question',
                               'val_ans': 'dev.span',
                               'vocab_file': 'vocab.dat',
                               'embedding': 'glove.trimmed.300.npz'}

    def download_model(self):
        # Validate contents of data_path folder:
        data_path = self.data_path
        download = False
        for file_name in self.file_name_dict.values():
            if not os.path.exists(os.path.join(data_path, file_name)):
                # prompt
                download = True
                print("The following required file is missing :", file_name)

        if download is True:
            if self.prompt is True:
                license_prompt('mrc_data',
                               'https://s3-us-west-2.amazonaws.com/nlp-architect-data/models/mrc'
                               '/mrc_data.zip',
                               self.data_dir)
                license_prompt('mrc_model',
                               'https://s3-us-west-2.amazonaws.com/nlp-architect-data/models/mrc'
                               '/mrc_model.zip',
                               self.model_dir)
            data_zipfile = os.path.join(self.data_dir, 'mrc_data.zip')
            model_zipfile = os.path.join(self.model_dir, 'mrc_model.zip')
            makedirs(self.data_dir, exist_ok=True)
            makedirs(self.model_dir, exist_ok=True)
            download_unlicensed_file('https://s3-us-west-2.amazonaws.com/nlp-architect-data'
                                     '/models/mrc/',
                                     'mrc_data.zip', data_zipfile)
            download_unlicensed_file('https://s3-us-west-2.amazonaws.com/nlp-architect-data'
                                     '/models/mrc/',
                                     'mrc_model.zip', model_zipfile)
            with zipfile.ZipFile(data_zipfile) as data_zip_ref:
                data_zip_ref.extractall(self.data_dir)
            with zipfile.ZipFile(model_zipfile) as model_zip_ref:
                model_zip_ref.extractall(self.model_dir)

    def load_model(self):
        select_device = 'GPU'
        restore_model = True
        # Create dictionary of filenames
        self.download_model()

        data_path = self.data_path
        # Paths for preprcessed files
        path_gen = data_path  # data is actually in mrc_data/data not, mrc_data
        train_para_ids = os.path.join(path_gen, self.file_name_dict['train_para_ids'])
        train_ques_ids = os.path.join(path_gen, self.file_name_dict['train_ques_ids'])
        answer_file = os.path.join(path_gen, self.file_name_dict['train_answer'])
        val_paras_ids = os.path.join(path_gen, self.file_name_dict['val_para_ids'])
        val_ques_ids = os.path.join(path_gen, self.file_name_dict['val_ques_ids'])
        val_ans_file = os.path.join(path_gen, self.file_name_dict['val_ans'])
        vocab_file = os.path.join(path_gen, self.file_name_dict['vocab_file'])

        model_dir = self.model_path
        # Create model dir if it doesn't exist
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        model_path = model_dir

        # Create lists for train and validation sets
        data_train = create_squad_training(train_para_ids, train_ques_ids, answer_file)
        data_dev = create_squad_training(val_paras_ids, val_ques_ids, val_ans_file)
        with open(vocab_file, encoding='UTF-8') as fp:
            vocab_list = fp.readlines()
        self.vocab_dict = {}
        self.vocab_rev = {}

        for i in range(len(vocab_list)):
            self.vocab_dict[i] = vocab_list[i].strip()
            self.vocab_rev[vocab_list[i].strip()] = i

            self.params_dict['train_set_size'] = len(data_train)

        # Combine train and dev data
        data_total = data_train + data_dev

        # obtain maximum length of question
        _, max_question = max_values_squad(data_total)
        self.params_dict['max_question'] = max_question

        # Load embeddings for vocab
        print('Loading Embeddings')
        embeddingz = np.load(os.path.join(path_gen, self.file_name_dict['embedding']))
        embeddings = embeddingz['glove']

        # Create train and dev sets
        print("Creating training and development sets")
        self.dev = get_data_array_squad(self.params_dict, data_dev, set_val='val')

        # Define Reading Comprehension model
        with tf.device('/device:' + select_device + ':0'):
            self.model = MatchLSTMAnswerPointer(self.params_dict, embeddings)

        # Define Configs for training
        run_config = tf.ConfigProto(allow_soft_placement=True, log_device_placement=False)

        # Create session run training
        self.sess = tf.Session(config=run_config)
        init = tf.global_variables_initializer()

        # Model Saver
        # pylint: disable=no-member
        model_saver = tf.train.Saver()
        model_ckpt = tf.train.get_checkpoint_state(model_path)
        idx_path = model_ckpt.model_checkpoint_path + ".index" if model_ckpt else ""

        # Initialize with random or pretrained weights
        # pylint: disable=no-member
        if model_ckpt and restore_model and (tf.gfile.Exists(
                model_ckpt.model_checkpoint_path) or tf.gfile.Exists(idx_path)):
            model_saver.restore(self.sess, model_ckpt.model_checkpoint_path)
            print("Loading from previously stored session")
        else:
            self.sess.run(init)

        shuffle(self.dev)

    @staticmethod
    def paragraphs(valid, vocab_tuple, num_examples):
        paragraphs = []
        vocab_forward = vocab_tuple[0]
        for idx in range(num_examples):
            test_paragraph = [vocab_forward[ele] for ele in valid[idx][0] if ele != 0]
            para_string = " ".join(map(str, test_paragraph))
            paragraphs.append(re.sub(r'\s([?.!,"](?:\s|$))', r'\1', para_string))  # (?:\s|$))
        return paragraphs

    @staticmethod
    def questions(valid, vocab_tuple, num_examples):
        vocab_forward = vocab_tuple[0]
        questions = []
        for idx in range(num_examples):
            test_question = [vocab_forward[ele] for ele in valid[idx][1] if ele != 0]
            ques_string = " ".join(map(str, test_question))
            questions.append(re.sub(r'\s([?.!"",])', r'\1', ques_string))
        return questions

    def inference(self, doc):
        body = doc
        print("Begin Inference Mode")
        question = body['question']
        paragraph_id = body['paragraph']
        return self.model.inference_mode(self.sess, self.dev, [self.vocab_dict, self.vocab_rev],
                                         dynamic_question_mode=True, num_examples=1, dropout=1.0,
                                         dynamic_usr_question=question,
                                         dynamic_question_index=paragraph_id)

    def get_paragraphs(self):
        ret = {'paragraphs': self.paragraphs(self.dev, [self.vocab_dict, self.vocab_rev],
                                             num_examples=5),
               'questions': self.questions(self.dev, [self.vocab_dict, self.vocab_rev],
                                           num_examples=5)}
        return ret



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# pylint: skip-file
"""GNMT attention sequence-to-sequence model with dynamic RNN support."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from nlp_architect.models.gnmt import attention_model
from nlp_architect.models.gnmt import model_helper
from nlp_architect.models.gnmt.utils import misc_utils as utils

__all__ = ["GNMTModel"]


class GNMTModel(attention_model.AttentionModel):
    """Sequence-to-sequence dynamic model with GNMT attention architecture
        with sparsity policy support.
    """

    def __init__(self,
                 hparams,
                 mode,
                 iterator,
                 source_vocab_table,
                 target_vocab_table,
                 reverse_target_vocab_table=None,
                 scope=None,
                 extra_args=None):
        self.is_gnmt_attention = (
            hparams.attention_architecture in ["gnmt", "gnmt_v2"])

        super(GNMTModel, self).__init__(
            hparams=hparams,
            mode=mode,
            iterator=iterator,
            source_vocab_table=source_vocab_table,
            target_vocab_table=target_vocab_table,
            reverse_target_vocab_table=reverse_target_vocab_table,
            scope=scope,
            extra_args=extra_args)

    def _build_encoder(self, hparams):
        """Build a GNMT encoder."""
        if hparams.encoder_type == "uni" or hparams.encoder_type == "bi":
            return super(GNMTModel, self)._build_encoder(hparams)

        if hparams.encoder_type != "gnmt":
            raise ValueError("Unknown encoder_type %s" % hparams.encoder_type)

        # Build GNMT encoder.
        num_bi_layers = 1
        num_uni_layers = self.num_encoder_layers - num_bi_layers
        utils.print_out("# Build a GNMT encoder")
        utils.print_out("  num_bi_layers = %d" % num_bi_layers)
        utils.print_out("  num_uni_layers = %d" % num_uni_layers)

        iterator = self.iterator
        source = iterator.source
        if self.time_major:
            source = tf.transpose(source)

        with tf.variable_scope("encoder") as scope:
            dtype = scope.dtype

            self.encoder_emb_inp = self.encoder_emb_lookup_fn(
                self.embedding_encoder, source)

            # Execute _build_bidirectional_rnn from Model class
            bi_encoder_outputs, bi_encoder_state = self._build_bidirectional_rnn(
                inputs=self.encoder_emb_inp,
                sequence_length=iterator.source_sequence_length,
                dtype=dtype,
                hparams=hparams,
                num_bi_layers=num_bi_layers,
                num_bi_residual_layers=0,  # no residual connection
            )

            # Build unidirectional layers
            if self.extract_encoder_layers:
                encoder_state, encoder_outputs = self._build_individual_encoder_layers(
                    bi_encoder_outputs, num_uni_layers, dtype, hparams)
            else:
                encoder_state, encoder_outputs = self._build_all_encoder_layers(
                    bi_encoder_outputs, num_uni_layers, dtype, hparams)

            # Pass all encoder states to the decoder
            #   except the first bi-directional layer
            encoder_state = (bi_encoder_state[1],) + (
                (encoder_state,) if num_uni_layers == 1 else encoder_state)

        return encoder_outputs, encoder_state

    def _build_all_encoder_layers(self, bi_encoder_outputs,
                                  num_uni_layers, dtype, hparams):
        """Build encoder layers all at once."""
        uni_cell = model_helper.create_rnn_cell(
            unit_type=hparams.unit_type,
            num_units=hparams.num_units,
            num_layers=num_uni_layers,
            num_residual_layers=self.num_encoder_residual_layers,
            forget_bias=hparams.forget_bias,
            dropout=hparams.dropout,
            num_gpus=self.num_gpus,
            base_gpu=1,
            mode=self.mode,
            single_cell_fn=self.single_cell_fn)
        encoder_outputs, encoder_state = tf.nn.dynamic_rnn(
            uni_cell,
            bi_encoder_outputs,
            dtype=dtype,
            sequence_length=self.iterator.source_sequence_length,
            time_major=self.time_major)

        # Use the top layer for now
        self.encoder_state_list = [encoder_outputs]

        return encoder_state, encoder_outputs

    def _build_individual_encoder_layers(self, bi_encoder_outputs,
                                         num_uni_layers, dtype, hparams):
        """Run each of the encoder layer separately, not used in general seq2seq."""
        uni_cell_lists = model_helper._cell_list(
            unit_type=hparams.unit_type,
            num_units=hparams.num_units,
            num_layers=num_uni_layers,
            num_residual_layers=self.num_encoder_residual_layers,
            forget_bias=hparams.forget_bias,
            dropout=hparams.dropout,
            num_gpus=self.num_gpus,
            base_gpu=1,
            mode=self.mode,
            single_cell_fn=self.single_cell_fn)

        encoder_inp = bi_encoder_outputs
        encoder_states = []
        self.encoder_state_list = [bi_encoder_outputs[:, :, :hparams.num_units],
                                   bi_encoder_outputs[:, :, hparams.num_units:]]
        with tf.variable_scope("rnn/multi_rnn_cell"):
            for i, cell in enumerate(uni_cell_lists):
                with tf.variable_scope("cell_%d" % i) as scope:
                    encoder_inp, encoder_state = tf.nn.dynamic_rnn(
                        cell,
                        encoder_inp,
                        dtype=dtype,
                        sequence_length=self.iterator.source_sequence_length,
                        time_major=self.time_major,
                        scope=scope)
                    encoder_states.append(encoder_state)
                    self.encoder_state_list.append(encoder_inp)

        encoder_state = tuple(encoder_states)
        encoder_outputs = self.encoder_state_list[-1]
        return encoder_state, encoder_outputs

    def _build_decoder_cell(self, hparams, encoder_outputs, encoder_state,
                            source_sequence_length):
        """Build a RNN cell with GNMT attention architecture."""
        # Standard attention
        if not self.is_gnmt_attention:
            return super(GNMTModel, self)._build_decoder_cell(
                hparams, encoder_outputs, encoder_state, source_sequence_length)

        # GNMT attention
        attention_option = hparams.attention
        attention_architecture = hparams.attention_architecture
        num_units = hparams.num_units
        infer_mode = hparams.infer_mode

        dtype = tf.float32

        if self.time_major:
            memory = tf.transpose(encoder_outputs, [1, 0, 2])
        else:
            memory = encoder_outputs

        if self.mode == tf.contrib.learn.ModeKeys.INFER and infer_mode == "beam_search":
            memory, source_sequence_length, encoder_state, batch_size = (
                self._prepare_beam_search_decoder_inputs(
                    hparams.beam_width, memory, source_sequence_length,
                    encoder_state))
        else:
            batch_size = self.batch_size

        attention_mechanism = self.attention_mechanism_fn(
            attention_option, num_units, memory, source_sequence_length, self.mode)

        cell_list = model_helper._cell_list(  # pylint: disable=protected-access
            unit_type=hparams.unit_type,
            num_units=num_units,
            num_layers=self.num_decoder_layers,
            num_residual_layers=self.num_decoder_residual_layers,
            forget_bias=hparams.forget_bias,
            dropout=hparams.dropout,
            num_gpus=self.num_gpus,
            mode=self.mode,
            single_cell_fn=self.single_cell_fn,
            residual_fn=gnmt_residual_fn
        )

        # Only wrap the bottom layer with the attention mechanism.
        attention_cell = cell_list.pop(0)

        # Only generate alignment in greedy INFER mode.
        alignment_history = (self.mode == tf.contrib.learn.ModeKeys.INFER
                             and infer_mode != "beam_search")
        attention_cell = tf.contrib.seq2seq.AttentionWrapper(
            attention_cell,
            attention_mechanism,
            attention_layer_size=None,  # don't use attention layer.
            output_attention=False,
            alignment_history=alignment_history,
            name="attention")

        if attention_architecture == "gnmt":
            cell = GNMTAttentionMultiCell(
                attention_cell, cell_list)
        elif attention_architecture == "gnmt_v2":
            cell = GNMTAttentionMultiCell(
                attention_cell, cell_list, use_new_attention=True)
        else:
            raise ValueError(
                "Unknown attention_architecture %s" % attention_architecture)

        if hparams.pass_hidden_state:
            decoder_initial_state = tuple(
                zs.clone(cell_state=es)
                if isinstance(zs, tf.contrib.seq2seq.AttentionWrapperState) else es
                for zs, es in zip(
                    cell.zero_state(batch_size, dtype), encoder_state))
        else:
            decoder_initial_state = cell.zero_state(batch_size, dtype)

        return cell, decoder_initial_state

    def _get_infer_summary(self, hparams):
        if hparams.infer_mode == "beam_search":
            return tf.no_op()
        elif self.is_gnmt_attention:
            return attention_model._create_attention_images_summary(
                self.final_context_state[0])
        else:
            return super(GNMTModel, self)._get_infer_summary(hparams)


class GNMTAttentionMultiCell(tf.nn.rnn_cell.MultiRNNCell):
    """A MultiCell with GNMT attention style."""

    def __init__(self, attention_cell, cells, use_new_attention=False):
        """Creates a GNMTAttentionMultiCell.

        Args:
          attention_cell: An instance of AttentionWrapper.
          cells: A list of RNNCell wrapped with AttentionInputWrapper.
          use_new_attention: Whether to use the attention generated from current
            step bottom layer's output. Default is False.
        """
        cells = [attention_cell] + cells
        self.use_new_attention = use_new_attention
        super(GNMTAttentionMultiCell, self).__init__(cells, state_is_tuple=True)

    def __call__(self, inputs, state, scope=None):
        """Run the cell with bottom layer's attention copied to all upper layers."""
        if not tf.contrib.framework.nest.is_sequence(state):
            raise ValueError(
                "Expected state to be a tuple of length %d, but received: %s"
                % (len(self.state_size), state))

        with tf.variable_scope(scope or "multi_rnn_cell"):
            new_states = []

            with tf.variable_scope("cell_0_attention"):
                attention_cell = self._cells[0]
                attention_state = state[0]
                cur_inp, new_attention_state = attention_cell(inputs, attention_state)
                new_states.append(new_attention_state)

            for i in range(1, len(self._cells)):
                with tf.variable_scope("cell_%d" % i):

                    cell = self._cells[i]
                    cur_state = state[i]

                    if self.use_new_attention:
                        cur_inp = tf.concat([cur_inp, new_attention_state.attention], -1)
                    else:
                        cur_inp = tf.concat([cur_inp, attention_state.attention], -1)

                    cur_inp, new_state = cell(cur_inp, cur_state)
                    new_states.append(new_state)

        return cur_inp, tuple(new_states)


def gnmt_residual_fn(inputs, outputs):
    """Residual function that handles different inputs and outputs inner dims.

    Args:
      inputs: cell inputs, this is actual inputs concatenated with the attention
        vector.
      outputs: cell outputs

    Returns:
      outputs + actual inputs
    """
    def split_input(inp, out):
        out_dim = out.get_shape().as_list()[-1]
        inp_dim = inp.get_shape().as_list()[-1]
        return tf.split(inp, [out_dim, inp_dim - out_dim], axis=-1)

    actual_inputs, _ = tf.contrib.framework.nest.map_structure(
        split_input, inputs, outputs)

    def assert_shape_match(inp, out):
        inp.get_shape().assert_is_compatible_with(out.get_shape())

    tf.contrib.framework.nest.assert_same_structure(actual_inputs, outputs)
    tf.contrib.framework.nest.map_structure(
        assert_shape_match, actual_inputs, outputs)
    return tf.contrib.framework.nest.map_structure(
        lambda inp, out: inp + out, actual_inputs, outputs)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""standard hparams utils."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def create_standard_hparams():
    return tf.contrib.training.HParams(
        # Data
        src="",
        tgt="",
        train_prefix="",
        dev_prefix="",
        test_prefix="",
        vocab_prefix="",
        embed_prefix="",
        out_dir="",

        # Networks
        num_units=512,
        num_encoder_layers=2,
        num_decoder_layers=2,
        dropout=0.2,
        unit_type="lstm",
        encoder_type="bi",
        residual=False,
        time_major=True,
        num_embeddings_partitions=0,
        num_enc_emb_partitions=0,
        num_dec_emb_partitions=0,

        # Attention mechanisms
        attention="scaled_luong",
        attention_architecture="standard",
        output_attention=True,
        pass_hidden_state=True,

        # Train
        optimizer="sgd",
        batch_size=128,
        init_op="uniform",
        init_weight=0.1,
        max_gradient_norm=5.0,
        learning_rate=1.0,
        warmup_steps=0,
        warmup_scheme="t2t",
        decay_scheme="luong234",
        colocate_gradients_with_ops=True,
        num_train_steps=12000,
        num_sampled_softmax=0,

        # Data constraints
        num_buckets=5,
        max_train=0,
        src_max_len=50,
        tgt_max_len=50,
        src_max_len_infer=0,
        tgt_max_len_infer=0,

        # Data format
        sos="<s>",
        eos="</s>",
        subword_option="",
        use_char_encode=False,
        check_special_token=True,

        # Misc
        forget_bias=1.0,
        num_gpus=1,
        epoch_step=0,  # record where we were within an epoch.
        steps_per_stats=100,
        steps_per_external_eval=0,
        share_vocab=False,
        metrics=["bleu"],
        log_device_placement=False,
        random_seed=None,
        # only enable beam search during inference when beam_width > 0.
        beam_width=0,
        length_penalty_weight=0.0,
        override_loaded_hparams=True,
        num_keep_ckpts=5,
        avg_ckpts=False,

        # For inference
        inference_indices=None,
        infer_batch_size=32,
        sampling_temperature=0.0,
        num_translations_per_input=1,
        infer_mode="greedy",

        # Language model
        language_model=False,
    )



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
#   pruning operations
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# pylint: skip-file
"""Basic sequence-to-sequence model with dynamic RNN support."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import collections

import tensorflow as tf
from tensorflow.contrib.model_pruning import get_masks, get_thresholds

from tensorflow.contrib.model_pruning.python.layers import core_layers
from tensorflow.contrib.model_pruning.python import pruning

from nlp_architect.models.gnmt import model_helper
from nlp_architect.models.gnmt.utils import iterator_utils
from nlp_architect.models.gnmt.utils import misc_utils as utils

utils.check_tensorflow_version()

__all__ = ["BaseModel", "Model"]


class TrainOutputTuple(collections.namedtuple(
    "TrainOutputTuple", ("train_summary", "train_loss", "predict_count",
                         "global_step", "word_count", "batch_size", "grad_norm",
                         "learning_rate", "mask_update_op", "pruning_summary"))):
    """To allow for flexibily in returing different outputs."""
    pass


class EvalOutputTuple(collections.namedtuple(
    "EvalOutputTuple", ("eval_loss", "predict_count",
                        "batch_size"))):
    """To allow for flexibily in returing different outputs."""
    pass


class InferOutputTuple(collections.namedtuple(
    "InferOutputTuple", ("infer_logits", "infer_summary", "sample_id",
                         "sample_words"))):
    """To allow for flexibily in returing different outputs."""
    pass


class BaseModel(object):
    """Sequence-to-sequence base class.
    """

    def __init__(self,
                 hparams,
                 mode,
                 iterator,
                 source_vocab_table,
                 target_vocab_table,
                 reverse_target_vocab_table=None,
                 scope=None,
                 extra_args=None):
        """Create the model.

        Args:
          hparams: Hyperparameter configurations.
          mode: TRAIN | EVAL | INFER
          iterator: Dataset Iterator that feeds data.
          source_vocab_table: Lookup table mapping source words to ids.
          target_vocab_table: Lookup table mapping target words to ids.
          reverse_target_vocab_table: Lookup table mapping ids to target words. Only
            required in INFER mode. Defaults to None.
          scope: scope of the model.
          extra_args: model_helper.ExtraArgs, for passing customizable functions.

        """
        # Set params
        self._set_params_initializer(hparams, mode, iterator,
                                     source_vocab_table, target_vocab_table,
                                     scope, extra_args)

        # Not used in general seq2seq models; when True, ignore decoder & training
        self.extract_encoder_layers = (hasattr(hparams, "extract_encoder_layers")
                                       and hparams.extract_encoder_layers)

        # Train graph
        res = self.build_graph(hparams, scope=scope)
        if not self.extract_encoder_layers:
            self._set_train_or_infer(res, reverse_target_vocab_table, hparams)

        # Saver
        self.saver = tf.train.Saver(
            tf.global_variables(), max_to_keep=hparams.num_keep_ckpts)

    def _set_params_initializer(self,
                                hparams,
                                mode,
                                iterator,
                                source_vocab_table,
                                target_vocab_table,
                                scope,
                                extra_args=None):
        """Set various params for self and initialize."""
        assert isinstance(iterator, iterator_utils.BatchedInput)
        self.iterator = iterator
        self.mode = mode
        self.src_vocab_table = source_vocab_table
        self.tgt_vocab_table = target_vocab_table

        self.src_vocab_size = hparams.src_vocab_size
        self.tgt_vocab_size = hparams.tgt_vocab_size
        self.num_gpus = hparams.num_gpus
        self.time_major = hparams.time_major

        if hparams.use_char_encode:
            assert (not self.time_major), ("Can't use time major for"
                                           " char-level inputs.")

        self.dtype = tf.float32
        self.num_sampled_softmax = hparams.num_sampled_softmax

        # extra_args: to make it flexible for adding external customizable code
        self.single_cell_fn = None
        if extra_args:
            self.single_cell_fn = extra_args.single_cell_fn

        # Set num units
        self.num_units = hparams.num_units

        # Set num layers
        self.num_encoder_layers = hparams.num_encoder_layers
        self.num_decoder_layers = hparams.num_decoder_layers
        assert self.num_encoder_layers
        assert self.num_decoder_layers

        # Set num residual layers
        if hasattr(hparams, "num_residual_layers"):  # compatible common_test_utils
            self.num_encoder_residual_layers = hparams.num_residual_layers
            self.num_decoder_residual_layers = hparams.num_residual_layers
        else:
            self.num_encoder_residual_layers = hparams.num_encoder_residual_layers
            self.num_decoder_residual_layers = hparams.num_decoder_residual_layers

        # Batch size
        self.batch_size = tf.size(self.iterator.source_sequence_length)

        # Global step
        self.global_step = tf.Variable(0, trainable=False)

        # Initializer
        self.random_seed = hparams.random_seed
        initializer = model_helper.get_initializer(
            hparams.init_op, self.random_seed, hparams.init_weight)
        tf.get_variable_scope().set_initializer(initializer)

        # Embeddings
        if extra_args and extra_args.encoder_emb_lookup_fn:
            self.encoder_emb_lookup_fn = extra_args.encoder_emb_lookup_fn
        else:
            self.encoder_emb_lookup_fn = tf.nn.embedding_lookup
        self.init_embeddings(hparams, scope)

    def _set_train_or_infer(self, res, reverse_target_vocab_table, hparams):
        """Set up training and inference."""
        if self.mode == tf.contrib.learn.ModeKeys.TRAIN:
            self.train_loss = res[1]
            self.word_count = tf.reduce_sum(
                self.iterator.source_sequence_length) + tf.reduce_sum(
                self.iterator.target_sequence_length)
        elif self.mode == tf.contrib.learn.ModeKeys.EVAL:
            self.eval_loss = res[1]
        elif self.mode == tf.contrib.learn.ModeKeys.INFER:
            self.infer_logits, _, self.final_context_state, self.sample_id = res
            self.sample_words = reverse_target_vocab_table.lookup(
                tf.to_int64(self.sample_id))

        if self.mode != tf.contrib.learn.ModeKeys.INFER:
            # Count the number of predicted words for compute ppl.
            self.predict_count = tf.reduce_sum(
                self.iterator.target_sequence_length)

        params = tf.trainable_variables()

        # Gradients and SGD update operation for training the model.
        # Arrange for the embedding vars to appear at the beginning.
        if self.mode == tf.contrib.learn.ModeKeys.TRAIN:
            self.learning_rate = tf.constant(hparams.learning_rate)
            # warm-up
            self.learning_rate = self._get_learning_rate_warmup(hparams)
            # decay
            self.learning_rate = self._get_learning_rate_decay(hparams)

            # Optimizer
            if hparams.optimizer == "sgd":
                opt = tf.train.GradientDescentOptimizer(self.learning_rate)
            elif hparams.optimizer == "adam":
                opt = tf.train.AdamOptimizer(self.learning_rate)
            else:
                raise ValueError("Unknown optimizer type %s" % hparams.optimizer)

            # Gradients
            gradients = tf.gradients(
                self.train_loss,
                params,
                colocate_gradients_with_ops=hparams.colocate_gradients_with_ops)

            clipped_grads, grad_norm_summary, grad_norm = model_helper.gradient_clip(
                gradients, max_gradient_norm=hparams.max_gradient_norm)
            self.grad_norm_summary = grad_norm_summary
            self.grad_norm = grad_norm

            self.update = opt.apply_gradients(
                zip(clipped_grads, params), global_step=self.global_step)

            # Summary
            self.train_summary = self._get_train_summary()
        elif self.mode == tf.contrib.learn.ModeKeys.INFER:
            self.infer_summary = self._get_infer_summary(hparams)

        # Print trainable variables
        utils.print_out("# Trainable variables")
        utils.print_out("Format: <name>, <shape>, <(soft) device placement>")
        for param in params:
            utils.print_out("  %s, %s, %s" % (param.name, str(param.get_shape()),
                                              param.op.device))

    def _get_learning_rate_warmup(self, hparams):
        """Get learning rate warmup."""
        warmup_steps = hparams.warmup_steps
        warmup_scheme = hparams.warmup_scheme
        utils.print_out("  learning_rate=%g, warmup_steps=%d, warmup_scheme=%s" %
                        (hparams.learning_rate, warmup_steps, warmup_scheme))

        # Apply inverse decay if global steps less than warmup steps.
        # Inspired by https://arxiv.org/pdf/1706.03762.pdf (Section 5.3)
        # When step < warmup_steps,
        #   learing_rate *= warmup_factor ** (warmup_steps - step)
        if warmup_scheme == "t2t":
            # 0.01^(1/warmup_steps): we start with a lr, 100 times smaller
            warmup_factor = tf.exp(tf.log(0.01) / warmup_steps)
            inv_decay = warmup_factor ** (
                tf.to_float(warmup_steps - self.global_step))
        else:
            raise ValueError("Unknown warmup scheme %s" % warmup_scheme)

        return tf.cond(
            self.global_step < hparams.warmup_steps,
            lambda: inv_decay * self.learning_rate,
            lambda: self.learning_rate,
            name="learning_rate_warump_cond")

    def _get_decay_info(self, hparams):
        """Return decay info based on decay_scheme."""
        if hparams.decay_scheme in ["luong5", "luong10", "luong234"]:
            decay_factor = 0.5
            if hparams.decay_scheme == "luong5":
                start_decay_step = int(hparams.num_train_steps / 2)
                decay_times = 5
            elif hparams.decay_scheme == "luong10":
                start_decay_step = int(hparams.num_train_steps / 2)
                decay_times = 10
            elif hparams.decay_scheme == "luong234":
                start_decay_step = int(hparams.num_train_steps * 2 / 3)
                decay_times = 4
            remain_steps = hparams.num_train_steps - start_decay_step
            decay_steps = int(remain_steps / decay_times)
        elif not hparams.decay_scheme:  # no decay
            start_decay_step = hparams.num_train_steps
            decay_steps = 0
            decay_factor = 1.0
        elif hparams.decay_scheme:
            raise ValueError("Unknown decay scheme %s" % hparams.decay_scheme)
        return start_decay_step, decay_steps, decay_factor

    def _get_learning_rate_decay(self, hparams):
        """Get learning rate decay."""
        start_decay_step, decay_steps, decay_factor = self._get_decay_info(hparams)
        utils.print_out("  decay_scheme=%s, start_decay_step=%d, decay_steps %d, "
                        "decay_factor %g" % (hparams.decay_scheme,
                                             start_decay_step,
                                             decay_steps,
                                             decay_factor))

        return tf.cond(
            self.global_step < start_decay_step,
            lambda: self.learning_rate,
            lambda: tf.train.exponential_decay(
                self.learning_rate,
                (self.global_step - start_decay_step),
                decay_steps, decay_factor, staircase=True),
            name="learning_rate_decay_cond")

    def init_embeddings(self, hparams, scope):
        """Init embeddings."""
        self.embedding_encoder, self.embedding_decoder = (
            model_helper.create_emb_for_encoder_and_decoder(
                share_vocab=hparams.share_vocab,
                src_vocab_size=self.src_vocab_size,
                tgt_vocab_size=self.tgt_vocab_size,
                src_embed_size=self.num_units,
                tgt_embed_size=self.num_units,
                num_enc_partitions=hparams.num_enc_emb_partitions,
                num_dec_partitions=hparams.num_dec_emb_partitions,
                src_vocab_file=hparams.src_vocab_file,
                tgt_vocab_file=hparams.tgt_vocab_file,
                src_embed_file=hparams.src_embed_file,
                tgt_embed_file=hparams.tgt_embed_file,
                use_char_encode=hparams.use_char_encode,
                scope=scope,
                embed_type=hparams.embedding_type, ))

    def _get_train_summary(self):
        """Get train summary."""
        train_summary = tf.summary.merge(
            [tf.summary.scalar("lr", self.learning_rate),
             tf.summary.scalar("train_loss", self.train_loss)] + self.grad_norm_summary)
        return train_summary

    def train(self, sess):
        """Execute train graph."""
        assert self.mode == tf.contrib.learn.ModeKeys.TRAIN
        output_tuple = TrainOutputTuple(train_summary=self.train_summary,
                                        train_loss=self.train_loss,
                                        predict_count=self.predict_count,
                                        global_step=self.global_step,
                                        word_count=self.word_count,
                                        batch_size=self.batch_size,
                                        grad_norm=self.grad_norm,
                                        learning_rate=self.learning_rate,
                                        mask_update_op=self.mask_update_op,
                                        pruning_summary=self.pruning_summary)
        return sess.run([self.update, output_tuple])

    def eval(self, sess):
        """Execute eval graph."""
        assert self.mode == tf.contrib.learn.ModeKeys.EVAL
        output_tuple = EvalOutputTuple(eval_loss=self.eval_loss,
                                       predict_count=self.predict_count,
                                       batch_size=self.batch_size)
        return sess.run(output_tuple)

    def build_graph(self, hparams, scope=None):
        """Subclass must implement this method.

        Creates a sequence-to-sequence model with dynamic RNN decoder API.
        Args:
          hparams: Hyperparameter configurations.
          scope: VariableScope for the created subgraph; default "dynamic_seq2seq".

        Returns:
          A tuple of the form (logits, loss_tuple, final_context_state, sample_id),
          where:
            logits: float32 Tensor [batch_size x num_decoder_symbols].
            loss: loss = the total loss / batch_size.
            final_context_state: the final state of decoder RNN.
            sample_id: sampling indices.

        Raises:
          ValueError: if encoder_type differs from mono and bi, or
            attention_option is not (luong | scaled_luong |
            bahdanau | normed_bahdanau).
        """
        utils.print_out("# Creating %s graph ..." % self.mode)

        # Projection
        if not self.extract_encoder_layers:
            with tf.variable_scope(scope or "build_network"):
                with tf.variable_scope("decoder/output_projection"):
                    if hparams.projection_type == 'sparse':
                        self.output_layer = core_layers.MaskedFullyConnected(
                            hparams.tgt_vocab_size, use_bias=False, name="output_projection")
                    elif hparams.projection_type == 'dense':
                        self.output_layer = tf.layers.Dense(
                            hparams.tgt_vocab_size, use_bias=False, name="output_projection")
                    else:
                        raise ValueError("Unknown projection type %s!" % hparams.projection_type)

        with tf.variable_scope(scope or "dynamic_seq2seq", dtype=self.dtype):
            # Encoder
            if hparams.language_model:  # no encoder for language modeling
                utils.print_out("  language modeling: no encoder")
                self.encoder_outputs = None
                encoder_state = None
            else:
                self.encoder_outputs, encoder_state = self._build_encoder(hparams)

            # Skip decoder if extracting only encoder layers
            if self.extract_encoder_layers:
                return

            # Decoder
            logits, decoder_cell_outputs, sample_id, final_context_state = (
                self._build_decoder(self.encoder_outputs, encoder_state, hparams))

            # Loss
            if self.mode != tf.contrib.learn.ModeKeys.INFER:
                with tf.device(model_helper.get_device_str(self.num_encoder_layers - 1,
                                                           self.num_gpus)):
                    loss = self._compute_loss(logits, decoder_cell_outputs)
            else:
                loss = tf.constant(0.0)

            # model pruning
            if hparams.pruning_hparams is not None:
                pruning_hparams = pruning.get_pruning_hparams().parse(
                    hparams.pruning_hparams)
                self.p = pruning.Pruning(pruning_hparams, global_step=self.global_step)
                self.mask_update_op = self.p.conditional_mask_update_op()
                masks = get_masks()
                thresholds = get_thresholds()
                masks_s = []
                for index, mask in enumerate(masks):
                    masks_s.append(tf.summary.scalar(mask.name + '/sparsity',
                                                     tf.nn.zero_fraction(mask)))
                    masks_s.append(tf.summary.scalar(thresholds[index].op.name + '/threshold',
                                                     thresholds[index]))
                    masks_s.append(tf.summary.histogram(mask.name + '/mask_tensor', mask))
                self.pruning_summary = tf.summary.merge(
                    [tf.summary.scalar('sparsity', self.p._sparsity),
                     tf.summary.scalar('last_mask_update_step', self.p._last_update_step)]
                    + masks_s)
            else:
                self.mask_update_op = tf.no_op()
                self.pruning_summary = tf.no_op()

            return logits, loss, final_context_state, sample_id

    @abc.abstractmethod
    def _build_encoder(self, hparams):
        """Subclass must implement this.

        Build and run an RNN encoder.

        Args:
          hparams: Hyperparameters configurations.

        Returns:
          A tuple of encoder_outputs and encoder_state.
        """
        pass

    def _build_encoder_cell(self, hparams, num_layers, num_residual_layers,
                            base_gpu=0):
        """Build a multi-layer RNN cell that can be used by encoder."""

        return model_helper.create_rnn_cell(
            unit_type=hparams.unit_type,
            num_units=self.num_units,
            num_layers=num_layers,
            num_residual_layers=num_residual_layers,
            forget_bias=hparams.forget_bias,
            dropout=hparams.dropout,
            num_gpus=hparams.num_gpus,
            mode=self.mode,
            base_gpu=base_gpu,
            single_cell_fn=self.single_cell_fn)

    def _get_infer_maximum_iterations(self, hparams, source_sequence_length):
        """Maximum decoding steps at inference time."""
        if hparams.tgt_max_len_infer:
            maximum_iterations = hparams.tgt_max_len_infer
            utils.print_out("  decoding maximum_iterations %d" % maximum_iterations)
        else:
            # TODO(thangluong): add decoding_length_factor flag
            decoding_length_factor = 2.0
            max_encoder_length = tf.reduce_max(source_sequence_length)
            maximum_iterations = tf.to_int32(tf.round(
                tf.to_float(max_encoder_length) * decoding_length_factor))
        return maximum_iterations

    def _build_decoder(self, encoder_outputs, encoder_state, hparams):
        """Build and run a RNN decoder with a final projection layer.

        Args:
          encoder_outputs: The outputs of encoder for every time step.
          encoder_state: The final state of the encoder.
          hparams: The Hyperparameters configurations.

        Returns:
          A tuple of final logits and final decoder state:
            logits: size [time, batch_size, vocab_size] when time_major=True.
        """
        tgt_sos_id = tf.cast(self.tgt_vocab_table.lookup(tf.constant(hparams.sos)),
                             tf.int32)
        tgt_eos_id = tf.cast(self.tgt_vocab_table.lookup(tf.constant(hparams.eos)),
                             tf.int32)
        iterator = self.iterator

        # maximum_iteration: The maximum decoding steps.
        maximum_iterations = self._get_infer_maximum_iterations(
            hparams, iterator.source_sequence_length)

        # Decoder.
        with tf.variable_scope("decoder") as decoder_scope:
            cell, decoder_initial_state = self._build_decoder_cell(
                hparams, encoder_outputs, encoder_state,
                iterator.source_sequence_length)

            # Optional ops depends on which mode we are in and which loss function we
            # are using.
            logits = tf.no_op()
            decoder_cell_outputs = None

            # Train or eval
            if self.mode != tf.contrib.learn.ModeKeys.INFER:
                # decoder_emp_inp: [max_time, batch_size, num_units]
                target_input = iterator.target_input
                if self.time_major:
                    target_input = tf.transpose(target_input)
                decoder_emb_inp = tf.nn.embedding_lookup(
                    self.embedding_decoder, target_input)

                # Helper
                helper = tf.contrib.seq2seq.TrainingHelper(
                    decoder_emb_inp, iterator.target_sequence_length,
                    time_major=self.time_major)

                # Decoder
                my_decoder = tf.contrib.seq2seq.BasicDecoder(
                    cell,
                    helper,
                    decoder_initial_state, )

                # Dynamic decoding
                outputs, final_context_state, _ = tf.contrib.seq2seq.dynamic_decode(
                    my_decoder,
                    output_time_major=self.time_major,
                    swap_memory=True,
                    scope=decoder_scope)

                sample_id = outputs.sample_id

                if self.num_sampled_softmax > 0:
                    # Note: this is required when using sampled_softmax_loss.
                    decoder_cell_outputs = outputs.rnn_output

                # Note: there's a subtle difference here between train and inference.
                # We could have set output_layer when create my_decoder
                #   and shared more code between train and inference.
                # We chose to apply the output_layer to all timesteps for speed:
                #   10% improvements for small models & 20% for larger ones.
                # If memory is a concern, we should apply output_layer per timestep.
                num_layers = self.num_decoder_layers
                num_gpus = self.num_gpus
                device_id = num_layers if num_layers < num_gpus else (num_layers - 1)
                # Colocate output layer with the last RNN cell if there is no extra GPU
                # available. Otherwise, put last layer on a separate GPU.
                with tf.device(model_helper.get_device_str(device_id, num_gpus)):
                    logits = self.output_layer(outputs.rnn_output)

                if self.num_sampled_softmax > 0:
                    logits = tf.no_op()  # unused when using sampled softmax loss.

            # Inference
            else:
                infer_mode = hparams.infer_mode
                start_tokens = tf.fill([self.batch_size], tgt_sos_id)
                end_token = tgt_eos_id
                utils.print_out(
                    "  decoder: infer_mode=%sbeam_width=%d, length_penalty=%f" % (
                        infer_mode, hparams.beam_width, hparams.length_penalty_weight))

                if infer_mode == "beam_search":
                    beam_width = hparams.beam_width
                    length_penalty_weight = hparams.length_penalty_weight

                    my_decoder = tf.contrib.seq2seq.BeamSearchDecoder(
                        cell=cell,
                        embedding=self.embedding_decoder,
                        start_tokens=start_tokens,
                        end_token=end_token,
                        initial_state=decoder_initial_state,
                        beam_width=beam_width,
                        output_layer=self.output_layer,
                        length_penalty_weight=length_penalty_weight)
                elif infer_mode == "sample":
                    # Helper
                    sampling_temperature = hparams.sampling_temperature
                    assert sampling_temperature > 0.0, (
                        "sampling_temperature must greater than 0.0 when using sample"
                        " decoder.")
                    helper = tf.contrib.seq2seq.SampleEmbeddingHelper(
                        self.embedding_decoder, start_tokens, end_token,
                        softmax_temperature=sampling_temperature,
                        seed=self.random_seed)
                elif infer_mode == "greedy":
                    helper = tf.contrib.seq2seq.GreedyEmbeddingHelper(
                        self.embedding_decoder, start_tokens, end_token)
                else:
                    raise ValueError("Unknown infer_mode '%s'", infer_mode)

                if infer_mode != "beam_search":
                    my_decoder = tf.contrib.seq2seq.BasicDecoder(
                        cell,
                        helper,
                        decoder_initial_state,
                        output_layer=self.output_layer  # applied per timestep
                    )

                # Dynamic decoding
                outputs, final_context_state, _ = tf.contrib.seq2seq.dynamic_decode(
                    my_decoder,
                    maximum_iterations=maximum_iterations,
                    output_time_major=self.time_major,
                    swap_memory=True,
                    scope=decoder_scope)

                if infer_mode == "beam_search":
                    sample_id = outputs.predicted_ids
                else:
                    logits = outputs.rnn_output
                    sample_id = outputs.sample_id

        return logits, decoder_cell_outputs, sample_id, final_context_state

    def get_max_time(self, tensor):
        time_axis = 0 if self.time_major else 1
        return tensor.shape[time_axis].value or tf.shape(tensor)[time_axis]

    @abc.abstractmethod
    def _build_decoder_cell(self, hparams, encoder_outputs, encoder_state,
                            source_sequence_length):
        """Subclass must implement this.

        Args:
          hparams: Hyperparameters configurations.
          encoder_outputs: The outputs of encoder for every time step.
          encoder_state: The final state of the encoder.
          source_sequence_length: sequence length of encoder_outputs.

        Returns:
          A tuple of a multi-layer RNN cell used by decoder and the intial state of
          the decoder RNN.
        """
        pass

    def _softmax_cross_entropy_loss(
            self, logits, decoder_cell_outputs, labels):
        """Compute softmax loss or sampled softmax loss."""
        if self.num_sampled_softmax > 0:

            is_sequence = (decoder_cell_outputs.shape.ndims == 3)

            if is_sequence:
                labels = tf.reshape(labels, [-1, 1])
                inputs = tf.reshape(decoder_cell_outputs, [-1, self.num_units])

            crossent = tf.nn.sampled_softmax_loss(
                weights=tf.transpose(self.output_layer.kernel),
                biases=self.output_layer.bias or tf.zeros([self.tgt_vocab_size]),
                labels=labels,
                inputs=inputs,
                num_sampled=self.num_sampled_softmax,
                num_classes=self.tgt_vocab_size,
                partition_strategy="div",
                seed=self.random_seed)

            if is_sequence:
                if self.time_major:
                    crossent = tf.reshape(crossent, [-1, self.batch_size])
                else:
                    crossent = tf.reshape(crossent, [self.batch_size, -1])

        else:
            crossent = tf.nn.sparse_softmax_cross_entropy_with_logits(
                labels=labels, logits=logits)

        return crossent

    def _compute_loss(self, logits, decoder_cell_outputs):
        """Compute optimization loss."""
        target_output = self.iterator.target_output
        if self.time_major:
            target_output = tf.transpose(target_output)
        max_time = self.get_max_time(target_output)

        crossent = self._softmax_cross_entropy_loss(
            logits, decoder_cell_outputs, target_output)

        target_weights = tf.sequence_mask(
            self.iterator.target_sequence_length, max_time, dtype=self.dtype)
        if self.time_major:
            target_weights = tf.transpose(target_weights)

        loss = tf.reduce_sum(
            crossent * target_weights) / tf.to_float(self.batch_size)
        return loss

    def _get_infer_summary(self, hparams):
        del hparams
        return tf.no_op()

    def infer(self, sess):
        assert self.mode == tf.contrib.learn.ModeKeys.INFER
        output_tuple = InferOutputTuple(infer_logits=self.infer_logits,
                                        infer_summary=self.infer_summary,
                                        sample_id=self.sample_id,
                                        sample_words=self.sample_words)
        return sess.run(output_tuple)

    def decode(self, sess):
        """Decode a batch.

        Args:
          sess: tensorflow session to use.

        Returns:
          A tuple consiting of outputs, infer_summary.
            outputs: of size [batch_size, time]
        """
        output_tuple = self.infer(sess)
        sample_words = output_tuple.sample_words
        infer_summary = output_tuple.infer_summary

        # make sure outputs is of shape [batch_size, time] or [beam_width,
        # batch_size, time] when using beam search.
        if self.time_major:
            sample_words = sample_words.transpose()
        elif sample_words.ndim == 3:
            # beam search output in [batch_size, time, beam_width] shape.
            sample_words = sample_words.transpose([2, 0, 1])
        return sample_words, infer_summary

    def build_encoder_states(self, include_embeddings=False):
        """Stack encoder states and return tensor [batch, length, layer, size]."""
        assert self.mode == tf.contrib.learn.ModeKeys.INFER
        if include_embeddings:
            stack_state_list = tf.stack(
                [self.encoder_emb_inp] + self.encoder_state_list, 2)
        else:
            stack_state_list = tf.stack(self.encoder_state_list, 2)

        # transform from [length, batch, ...] -> [batch, length, ...]
        if self.time_major:
            stack_state_list = tf.transpose(stack_state_list, [1, 0, 2, 3])

        return stack_state_list


class Model(BaseModel):
    """Sequence-to-sequence dynamic model.

    This class implements a multi-layer recurrent neural network as encoder,
    and a multi-layer recurrent neural network decoder.
    """

    def _build_encoder_from_sequence(self, hparams, sequence, sequence_length):
        """Build an encoder from a sequence.

        Args:
          hparams: hyperparameters.
          sequence: tensor with input sequence data.
          sequence_length: tensor with length of the input sequence.

        Returns:
          encoder_outputs: RNN encoder outputs.
          encoder_state: RNN encoder state.

        Raises:
          ValueError: if encoder_type is neither "uni" nor "bi".
        """
        num_layers = self.num_encoder_layers
        num_residual_layers = self.num_encoder_residual_layers

        if self.time_major:
            sequence = tf.transpose(sequence)

        with tf.variable_scope("encoder") as scope:
            dtype = scope.dtype

            self.encoder_emb_inp = self.encoder_emb_lookup_fn(
                self.embedding_encoder, sequence)

            # Encoder_outputs: [max_time, batch_size, num_units]
            if hparams.encoder_type == "uni":
                utils.print_out("  num_layers = %d, num_residual_layers=%d" %
                                (num_layers, num_residual_layers))
                cell = self._build_encoder_cell(hparams, num_layers,
                                                num_residual_layers)

                encoder_outputs, encoder_state = tf.nn.dynamic_rnn(
                    cell,
                    self.encoder_emb_inp,
                    dtype=dtype,
                    sequence_length=sequence_length,
                    time_major=self.time_major,
                    swap_memory=True)
            elif hparams.encoder_type == "bi":
                num_bi_layers = int(num_layers / 2)
                num_bi_residual_layers = int(num_residual_layers / 2)
                utils.print_out("  num_bi_layers = %d, num_bi_residual_layers=%d" %
                                (num_bi_layers, num_bi_residual_layers))

                encoder_outputs, bi_encoder_state = (
                    self._build_bidirectional_rnn(
                        inputs=self.encoder_emb_inp,
                        sequence_length=sequence_length,
                        dtype=dtype,
                        hparams=hparams,
                        num_bi_layers=num_bi_layers,
                        num_bi_residual_layers=num_bi_residual_layers))

                if num_bi_layers == 1:
                    encoder_state = bi_encoder_state
                else:
                    # alternatively concat forward and backward states
                    encoder_state = []
                    for layer_id in range(num_bi_layers):
                        encoder_state.append(bi_encoder_state[0][layer_id])  # forward
                        encoder_state.append(bi_encoder_state[1][layer_id])  # backward
                    encoder_state = tuple(encoder_state)
            else:
                raise ValueError("Unknown encoder_type %s" % hparams.encoder_type)

        # Use the top layer for now
        self.encoder_state_list = [encoder_outputs]

        return encoder_outputs, encoder_state

    def _build_encoder(self, hparams):
        """Build encoder from source."""
        utils.print_out("# Build a basic encoder")
        return self._build_encoder_from_sequence(
            hparams, self.iterator.source, self.iterator.source_sequence_length)

    def _build_bidirectional_rnn(self, inputs, sequence_length,
                                 dtype, hparams,
                                 num_bi_layers,
                                 num_bi_residual_layers,
                                 base_gpu=0):
        """Create and call biddirectional RNN cells.

        Args:
          num_residual_layers: Number of residual layers from top to bottom. For
            example, if `num_bi_layers=4` and `num_residual_layers=2`, the last 2 RNN
            layers in each RNN cell will be wrapped with `ResidualWrapper`.
          base_gpu: The gpu device id to use for the first forward RNN layer. The
            i-th forward RNN layer will use `(base_gpu + i) % num_gpus` as its
            device id. The `base_gpu` for backward RNN cell is `(base_gpu +
            num_bi_layers)`.

        Returns:
          The concatenated bidirectional output and the bidirectional RNN cell"s
          state.
        """
        # Construct forward and backward cells
        fw_cell = self._build_encoder_cell(hparams,
                                           num_bi_layers,
                                           num_bi_residual_layers,
                                           base_gpu=base_gpu)
        bw_cell = self._build_encoder_cell(hparams,
                                           num_bi_layers,
                                           num_bi_residual_layers,
                                           base_gpu=(base_gpu + num_bi_layers))

        bi_outputs, bi_state = tf.nn.bidirectional_dynamic_rnn(
            fw_cell,
            bw_cell,
            inputs,
            dtype=dtype,
            sequence_length=sequence_length,
            time_major=self.time_major,
            swap_memory=True)

        return tf.concat(bi_outputs, -1), bi_state

    def _build_decoder_cell(self, hparams, encoder_outputs, encoder_state,
                            source_sequence_length, base_gpu=0):
        """Build an RNN cell that can be used by decoder."""
        # We only make use of encoder_outputs in attention-based models
        if hparams.attention:
            raise ValueError("BasicModel doesn't support attention.")

        cell = model_helper.create_rnn_cell(
            unit_type=hparams.unit_type,
            num_units=self.num_units,
            num_layers=self.num_decoder_layers,
            num_residual_layers=self.num_decoder_residual_layers,
            forget_bias=hparams.forget_bias,
            dropout=hparams.dropout,
            num_gpus=self.num_gpus,
            mode=self.mode,
            single_cell_fn=self.single_cell_fn,
            base_gpu=base_gpu
        )

        if hparams.language_model:
            encoder_state = cell.zero_state(self.batch_size, self.dtype)
        elif not hparams.pass_hidden_state:
            raise ValueError("For non-attentional model, "
                             "pass_hidden_state needs to be set to True")

        # For beam search, we need to replicate encoder infos beam_width times
        if self.mode == tf.contrib.learn.ModeKeys.INFER and hparams.infer_mode == "beam_search":
            decoder_initial_state = tf.contrib.seq2seq.tile_batch(
                encoder_state, multiplier=hparams.beam_width)
        else:
            decoder_initial_state = encoder_state

        return cell, decoder_initial_state



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import copy
import re
import sys
from os import PathLike

from tqdm import tqdm

from nlp_architect.models.absa import TRAIN_OUT
from nlp_architect.models.absa.inference.data_types import Polarity
from nlp_architect.models.absa.train.data_types import AspectTerm, \
    DepRelation, DepRelationTerm, LoadOpinionStopLists, LoadAspectStopLists, OpinionTerm, \
    QualifiedTerm
from nlp_architect.models.absa import TRAIN_LEXICONS
from nlp_architect.models.absa.utils import _load_parsed_docs_from_dir, _write_final_lex, \
    _load_lex_as_list_from_csv, read_generic_lex_from_file
from nlp_architect.models.absa.train.rules import rule_1, rule_2, rule_3, rule_4, rule_5, rule_6


class AcquireTerms(object):
    """
    Lexicon acquisition. produce opinion lexicon and an aspect lexicon based
    on input dataset.

    Attributes:
        opinion_candidate_list_curr_iter (dict): candidate opinion terms in the current iteration
        opinion_candidate_list_prev_iter (dict): opinion candidates list of previous iteration
        opinion_candidate_list (dict): opinion terms learned across all iterations
        opinion_candidates_list_final (list): final opinion candidates list
        opinion_candidate_list_raw (dict): all instances of candidate opinion terms
                                           across all iterations
        aspect_candidate_list_curr_iter (dict): candidate terms in the current iteration
        aspects_candidate_list_prev_iter(list): Aspect candidates list of previous iteration
        aspect_candidate_list (list):  aspect terms learned across all iterations
        aspect_candidates_list_final (list): final aspect candidates list
        aspect_candidate_list_raw (dict): all instances of candidate aspect terms
                                          across all iterations
        """

    out_dir = TRAIN_OUT / 'output'
    feature_table_path = out_dir / 'feature_table.csv'
    generic_opinion_lex_path = TRAIN_LEXICONS / 'GenericOpinionLex.csv'
    acquired_opinion_terms_path = out_dir / 'generated_opinion_lex.csv'
    acquired_aspect_terms_path = out_dir / 'generated_aspect_lex.csv'

    GENERIC_OPINION_LEX = _load_lex_as_list_from_csv('GenericOpinionLex.csv')
    GENERAL_ADJECTIVES_LEX = _load_lex_as_list_from_csv('GeneralAdjectivesLex.csv')
    GENERIC_QUANTIFIERS_LEX = _load_lex_as_list_from_csv('GenericQuantifiersLex.csv')
    GEOGRAPHICAL_ADJECTIVES_LEX = _load_lex_as_list_from_csv('GeographicalAdjectivesLex.csv')
    INTENSIFIERS_LEX = _load_lex_as_list_from_csv('IntensifiersLex.csv')
    TIME_ADJECTIVE_LEX = _load_lex_as_list_from_csv('TimeAdjectiveLex.csv')
    ORDINAL_NUMBERS_LEX = _load_lex_as_list_from_csv('OrdinalNumbersLex.csv')
    PREPOSITIONS_LEX = _load_lex_as_list_from_csv('PrepositionsLex.csv')
    PRONOUNS_LEX = _load_lex_as_list_from_csv('PronounsLex.csv')
    COLORS_LEX = _load_lex_as_list_from_csv('ColorsLex.csv')
    DETERMINERS_LEX = _load_lex_as_list_from_csv('DeterminersLex.csv')
    NEGATION_LEX = _load_lex_as_list_from_csv('NegationLex.csv')

    OPINION_STOP_LIST = LoadOpinionStopLists(DETERMINERS_LEX,
                                             GENERAL_ADJECTIVES_LEX,
                                             GENERIC_QUANTIFIERS_LEX,
                                             GEOGRAPHICAL_ADJECTIVES_LEX,
                                             INTENSIFIERS_LEX, TIME_ADJECTIVE_LEX,
                                             ORDINAL_NUMBERS_LEX,
                                             PREPOSITIONS_LEX, COLORS_LEX, NEGATION_LEX)

    ASPECT_STOP_LIST = LoadAspectStopLists(GENERIC_OPINION_LEX,
                                           DETERMINERS_LEX,
                                           GENERAL_ADJECTIVES_LEX,
                                           GENERIC_QUANTIFIERS_LEX,
                                           GEOGRAPHICAL_ADJECTIVES_LEX,
                                           INTENSIFIERS_LEX, TIME_ADJECTIVE_LEX,
                                           ORDINAL_NUMBERS_LEX,
                                           PREPOSITIONS_LEX, PRONOUNS_LEX, COLORS_LEX,
                                           NEGATION_LEX)

    FILTER_PATTERNS = [re.compile(r'.*\d+.*')]
    FLOAT_FORMAT = '{0:.3g}'
    # maximum number of iterations
    MAX_NUM_OF_ITERATIONS = 3
    NUM_OF_SENTENCES_PER_OPINION_AND_ASPECT_TERM_INC = 35000

    opinion_candidate_list_prev_iter = read_generic_lex_from_file(generic_opinion_lex_path)
    generic_sent_dict = copy.deepcopy(opinion_candidate_list_prev_iter)
    opinion_candidate_list = {}
    opinion_candidate_list_raw = {}
    opinion_candidate_list_curr_iter = {}
    opinion_candidates_list_final = []
    aspect_candidate_list_raw = {}
    aspect_candidate_list = list()
    aspect_candidate_list_curr_iter = {}
    aspect_candidates_list_final = []
    init_aspect_dict = list()
    aspects_candidate_list_prev_iter = list()
    min_freq_opinion_candidate = 2
    min_freq_aspect_candidate = 3

    def extract_terms_from_doc(self, parsed_doc):
        """Extract candidate terms for sentences in parsed document.

        Args:
            parsed_doc (ParsedDocument): Input parsed document.
        """
        for text, parsed_sent in parsed_doc.sent_iter():
            relations = _get_rel_list(parsed_sent)

            for rel_entry in relations:
                if rel_entry.rel != 'root':
                    gov_seen = self.opinion_candidate_list_prev_iter.get(rel_entry.gov.text)
                    dep_seen = self.opinion_candidate_list_prev_iter.get(rel_entry.dep.text)
                    opinions = []
                    aspects = []

                    # =========================== acquisition rules ==============================

                    if bool(gov_seen) ^ bool(dep_seen):
                        opinions.append(rule_1(rel_entry, gov_seen, dep_seen, text))

                    if not gov_seen and dep_seen:
                        opinions.append(rule_2(rel_entry, relations, dep_seen, text))

                        aspects.append(rule_3(rel_entry, relations, text))

                        aspects.append(rule_4(rel_entry, relations, text))

                    if self.aspects_candidate_list_prev_iter and \
                            AspectTerm.from_token(rel_entry.gov) \
                            in self.aspects_candidate_list_prev_iter and \
                            AspectTerm.from_token(rel_entry.dep) \
                            not in self.aspects_candidate_list_prev_iter:

                        opinions.append(rule_5(rel_entry, text))
                        aspects.append(rule_6(rel_entry, relations, text))

                    self._add_opinion_term(opinions)
                    self._add_aspect_term(aspects)

    def extract_opinion_and_aspect_terms(self, parsed_document_iter, num_of_docs):
        """Extract candidate terms from parsed document iterator.

        Args:
            parsed_document_iter (Iterator): Parsed document iterator.
            num_of_docs (int): number of documents on iterator.
        """

        for parsed_document in tqdm(parsed_document_iter, total=num_of_docs, file=sys.stdout):
            self.extract_terms_from_doc(parsed_document)

    def _is_valid_term(self, cand_term):
        """Validates a candidate term.

        Args:
            cand_term (CandidateTerm): candidate terms list.
        """
        term = str(cand_term)
        for pattern in self.FILTER_PATTERNS:
            if pattern.match(term):
                return False
        if self.OPINION_STOP_LIST.is_in_stop_list(term):
            return False
        if term.lower() != term and term.upper() != term:
            return False
        return True

    def _add_aspect_term(self, terms):
        """
        add new aspect term to table.
        Args:
            terms (list of CandidateTerm): candidate terms list
        """
        for term in terms:
            if term:
                term_entry = AspectTerm(term.term, term.pos)
                if term_entry not in self.init_aspect_dict and \
                        term_entry not in self.aspect_candidate_list and not\
                        self.ASPECT_STOP_LIST.is_in_stop_list(term.term[0]):
                    _insert_new_term_to_table(term, self.aspect_candidate_list_curr_iter)

        return True

    def _add_opinion_term(self, terms):
        """
        Add new opinion term to table
        Args:
            terms (list of CandidateTerm): candidate term
        """
        for term in terms:
            if term and self._is_valid_term(term):
                if str(term.term[0]) not in self.generic_sent_dict.keys():
                    if str(term.term[0]) not in self.opinion_candidate_list:
                        _insert_new_term_to_table(term, self.opinion_candidate_list_curr_iter)

    def _insert_new_terms_to_tables(self):
        """
        Insert new terms to tables
        clear candidates lists from previous iteration

        """
        self.opinion_candidate_list_prev_iter = {}
        self.opinion_candidate_list_raw = _merge_tables(self.opinion_candidate_list_raw,
                                                        self.opinion_candidate_list_curr_iter)
        for cand_term_list in self.opinion_candidate_list_curr_iter.values():
            if len(cand_term_list) >= \
                    self.min_freq_opinion_candidate:
                new_opinion_term = _set_opinion_term_polarity(cand_term_list)
                self.opinion_candidate_list_prev_iter[
                    str(new_opinion_term)] = new_opinion_term
        self.opinion_candidate_list_curr_iter = {}
        self.opinion_candidate_list = {**self.opinion_candidate_list,
                                       **self.opinion_candidate_list_prev_iter}
        self.aspects_candidate_list_prev_iter = list()
        self.aspect_candidate_list_raw = _merge_tables(
            self.aspect_candidate_list_raw,
            self.aspect_candidate_list_curr_iter)
        for extracted_aspect_list in self.aspect_candidate_list_curr_iter.values():
            if len(extracted_aspect_list) >= \
                    self.min_freq_aspect_candidate:
                first = extracted_aspect_list[0]
                new_aspect_entry = AspectTerm(first.term, first.pos)
                if new_aspect_entry not in self.aspects_candidate_list_prev_iter:
                    self.aspects_candidate_list_prev_iter.append(new_aspect_entry)
        self.aspect_candidate_list_curr_iter = {}
        self.aspect_candidate_list = \
            self.aspect_candidate_list + self.aspects_candidate_list_prev_iter

    def _write_output(self):
        """
        write generated lexicons to csv files
        """
        out = AcquireTerms.out_dir
        out.mkdir(parents=True, exist_ok=True)

        _write_final_lex(self.opinion_candidates_list_final, self.acquired_opinion_terms_path)
        _write_final_lex(self.aspect_candidates_list_final, self.acquired_aspect_terms_path)

    def acquire_lexicons(self, parsed_dir: str or PathLike):
        """Acquire new opinion and aspect lexicons.

        Args:
            parsed_dir (PathLike): Path to parsed documents folder.
        """
        parsed_docs = _load_parsed_docs_from_dir(parsed_dir)
        dataset_sentence_len = 0
        for parsed_doc in parsed_docs.values():
            dataset_sentence_len += len(parsed_doc.sentences)

        add_to_thresholds = \
            int(dataset_sentence_len / self.NUM_OF_SENTENCES_PER_OPINION_AND_ASPECT_TERM_INC)
        self.min_freq_opinion_candidate += add_to_thresholds
        self.min_freq_aspect_candidate += add_to_thresholds

        for iteration_num in range(self.MAX_NUM_OF_ITERATIONS):
            if len(self.opinion_candidate_list_prev_iter) == 0 \
                    and len(self.aspects_candidate_list_prev_iter) == 0:
                break

            print("\n#Iteration: {}".format(iteration_num + 1))

            self.extract_opinion_and_aspect_terms(iter(parsed_docs.values()),
                                                  len(parsed_docs))

            self._insert_new_terms_to_tables()

        self.opinion_candidates_list_final = \
            generate_final_opinion_candidates_list(
                self.opinion_candidate_list_raw, self.opinion_candidates_list_final,
                self.min_freq_opinion_candidate)
        self.aspect_candidates_list_final = \
            _generate_final_aspect_candidates_list(
                self.aspect_candidate_list_raw,
                self.aspect_candidates_list_final,
                self.min_freq_aspect_candidate)

        self._write_output()

        aspect_dict = {}
        for cand_term in self.aspect_candidates_list_final:
            aspect_dict[cand_term.term[0]] = cand_term.frequency

        return aspect_dict


def _get_rel_list(parsed_sentence):
    res = []
    gen_toks = []
    for tok in parsed_sentence:
        gen_toks.append(
            DepRelationTerm(tok['text'], tok['lemma'], tok['pos'], tok['ner'], tok['start']))

    for gen_tok, tok in zip(gen_toks, parsed_sentence):
        gov_idx = tok['gov']
        if gov_idx != -1:
            res.append(DepRelation(gen_toks[gov_idx], gen_tok, tok['rel']))
    return res


def _merge_tables(d1, d2):
    """
    Merge dictionaries
    Args:
        d1 (dict): first dict to merge
        d2 (dict): second dict to merge
    """
    for key, l in d2.items():
        if key in d1:
            for item in l:
                if item not in d1[key]:
                    d1[key].append(item)
        else:
            d1[key] = l
    return d1


def _insert_new_term_to_table(term, curr_table):
    """
    Insert term to table of lists.
    Args:
        term (term): term to be inserted
        curr_table (dict): input table
    """
    table_key_word = str(term)
    if table_key_word:
        if table_key_word in curr_table and term not in curr_table[table_key_word]:
            curr_table[table_key_word].append(term)
        else:
            curr_table[table_key_word] = [term]


def _set_opinion_term_polarity(terms_list):
    """Set opinion term polarity.

    Args:
        terms_list (list): list of opinion terms
    """
    first = terms_list[0]
    new_term = first.term

    positive_pol = 0
    negative_pol = 0
    pol = None
    for term in terms_list:
        try:
            pol = term.term_polarity
        except Exception as e:
            print("extracted_term missing term_polarity: " + str(e))
        if pol is not None:
            if pol == Polarity.POS:
                positive_pol = positive_pol + 1
            if pol == Polarity.NEG:
                negative_pol = negative_pol + 1
    new_term_polarity = Polarity.UNK
    if positive_pol >= negative_pol and positive_pol > 0:
        new_term_polarity = Polarity.POS
    elif negative_pol >= positive_pol and negative_pol > 0:
        new_term_polarity = Polarity.NEG

    return OpinionTerm(new_term, new_term_polarity)


def _generate_final_aspect_candidates_list(aspect_candidate_list_raw,
                                           final_aspect_candidates_list,
                                           frequency_threshold):
    """
    generate final aspect candidates list from map
    Args:
        aspect_candidate_list_raw (dict): key = term, value =
        lists of candidate terms.
        final_aspect_candidates_list (list): list of final aspect candidates
        frequency_threshold (int): minimum freq. for qualifying term
    """
    term_polarity = Polarity.UNK
    for extracted_term_list in aspect_candidate_list_raw.values():
        if len(extracted_term_list) >= frequency_threshold:
            term = extracted_term_list[0]
            qualified_term = QualifiedTerm(term.term, term.pos,
                                           len(extracted_term_list),
                                           term_polarity)
            final_aspect_candidates_list.append(qualified_term)

    return final_aspect_candidates_list


def generate_final_opinion_candidates_list(opinion_candidate_list_raw,
                                           final_opinion_candidates_list,
                                           frequency_threshold):
    """
    generate final opinion candidates list from raw opinion candidate list
    Args:
        opinion_candidate_list_raw (dict): key = term, value =
        lists of extracted terms.
        final_opinion_candidates_list (list): list of final opinion candidates
        frequency_threshold (int): minimum freq. for qualifying term
    """
    for candidate_list in opinion_candidate_list_raw.values():
        positive_pol = 0
        negative_pol = 0
        if len(candidate_list) >= frequency_threshold:
            for candidate in candidate_list:
                pol = candidate.term_polarity
                if pol is not None:
                    if pol == Polarity.POS:
                        positive_pol = positive_pol + 1
                    if pol == Polarity.NEG:
                        negative_pol = negative_pol + 1

            term_polarity = Polarity.UNK
            if positive_pol > negative_pol and positive_pol > 0:
                term_polarity = Polarity.POS
            elif negative_pol >= positive_pol and negative_pol > 0:
                term_polarity = Polarity.NEG

            term = candidate_list[0]

            qualified_term = QualifiedTerm(term.term, term.pos,
                                           len(candidate_list), term_polarity)
            final_opinion_candidates_list.append(qualified_term)

    return final_opinion_candidates_list



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import division, print_function, unicode_literals, absolute_import

from seqeval.metrics import classification_report


def get_conll_scores(predictions, y, y_lex, unk='O'):
    if isinstance(predictions, list):
        predictions = predictions[-1]
    test_p = predictions
    if len(test_p.shape) > 2:
        test_p = test_p.argmax(2)
    test_y = y
    if len(test_y.shape) > 2:
        test_y = test_y.argmax(2)

    prediction_data = []
    for n in range(test_y.shape[0]):
        test_yval = []
        for i in list(test_y[n]):
            try:
                test_yval.append(y_lex[i])
            except KeyError:
                pass
        test_pval = [unk] * len(test_yval)
        for e, i in enumerate(list(test_p[n])[:len(test_pval)]):
            try:
                test_pval[e] = y_lex[i]
            except KeyError:
                pass
        prediction_data.append((test_yval, test_pval))
    y_true, y_pred = list(zip(*prediction_data))
    return classification_report(y_true, y_pred, digits=3)
